"""
This module contains methods related to the Fractions Skill Score (FSS).

For an explanation of the FSS, and implementation considerations, see references below.

Currently uses the default score defined Robert and Leans (2008) :sup:`[1,2]`. Fine grained controls are considered in `#353 <GITHUB353_>`_.

The default computation is performed using`numpy` :sup:`[3]` and summed-area tables, with future optimization options considered in:
- `#269 <GITHUB269_>`_.
- `#270 <GITHUB270_>`_.

References:
1. Roberts, N. M., and H. W. Lean, 2008: Scale-Selective Verification of Rainfall Accumulations from
   High-Resolution Forecasts of Convective Events. Monthly Weather Review, 136, 78–97,
   https://doi.org/10.1175/2007mwr2123.1.
2. Mittermaier, M. P., 2021: A “Meta” Analysis of the Fractions Skill Score: The Limiting Case and
   Implications for Aggregation. Monthly Weather Review, 149, 3491–3504,
   https://doi.org/10.1175/mwr-d-18-0106.1.
3. FAGGIAN, N., B. ROUX, P. STEINLE, and B. EBERT, 2015: Fast calculation of the fractions skill
   score. MAUSAM, 66, 457–466, https://doi.org/10.54302/mausam.v66i3.555.

.. _GITHUB269: https://github.com/nci/scores/issues/269
.. _GITHUB270: https://github.com/nci/scores/issues/270
.. _GITHUB353: https://github.com/nci/scores/issues/353
"""
# sphinx docstrings: issue with inconsistent type rerpesentations
# see: https://github.com/sphinx-doc/sphinx/issues/9813
from __future__ import annotations  # isort: off

import warnings
from typing import Callable, Optional, Tuple

import numpy as np
import xarray as xr
from numpy import typing as npt

from scores.fast.fss.fss_backends import get_compute_backend
from scores.fast.fss.typing import FssComputeMethod, FssDecomposed
from scores.typing import FlexibleDimensionTypes
from scores.utils import (
    DimensionError,
    DimensionWarning,
    FieldTypeError,
    NumpyThresholdOperator,
    gather_dimensions,
    left_identity_operator,
)


def fss_2d(  # pylint: disable=too-many-locals,too-many-arguments
    fcst: xr.DataArray,
    obs: xr.DataArray,
    *,  # Force keywords arguments to be keyword-only
    event_threshold: float,
    window_size: Tuple[int, int],
    spatial_dims: Tuple[str, str],
    zero_padding: bool = False,
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    threshold_operator: Optional[Callable] = None,
    compute_method: FssComputeMethod = FssComputeMethod.NUMPY,
    dask: str = "forbidden",  # see: `xarray.apply_ufunc` for options
) -> xr.DataArray:
    """
    Uses :py:func:`fss_2d_single_field` to compute the Fractions Skill Score (FSS) for each 2D
    spatial field in the DataArray and then aggregates them over the output of gather dimensions.

    The aggregation method is the intended extension of the score defined by Robert and Leans (2008)
    for multiple forecasts :sup:`[1,2]`

    .. note::
        This method takes in a ``threshold_operator`` to compare the input
        fields against the ``event_threshold`` which defaults to ``numpy.greater``,
        and is compatible with lightweight numpy operators. If you would
        like to do more advanced binary discretization consider doing this
        separately and using :py:func:`scores.spatial.fss_2d_binary` instead, which takes in a
        pre-discretized binary field.

    Optionally aggregates the output along other dimensions if `reduce_dims` or
    (mutually exclusive) ``preserve_dims`` are specified.

    For implementation for a single 2-D field,
    see: :py:func:`fss_2d_single_field`. This offers the user the ability to use
    their own aggregation methods, rather than ``xarray.ufunc``

    .. warning::
        ``dask`` option is not fully tested

    Args:
        fcst: An array of forecasts
        obs: An array of observations (same spatial shape as ``fcst``)
        event_threshold: A scalar to compare ``fcst`` and ``obs`` fields to generate a
            binary "event" field. (defaults to ``np.greater``).
        window_size: A pair of positive integers ``(height, width)`` of the sliding
            window_size; the window size must be greater than 0 and fit within
            the shape of ``obs`` and ``fcst``.
        spatial_dims: A pair of dimension names ``(x, y)`` where ``x`` and ``y`` are
            the spatial dimensions to slide the window along to compute the FSS.
            e.g. ``("lat", "lon")``.
        zero_padding: If set to true, applies a 0-valued window border around the
            data field before computing the FSS. If set to false (default), it
            uses the edges (thickness = window size) of the input fields as the border.
            One can think of it as:
            - zero_padding = False => inner border
            - zero_padding = True => outer border with 0 values
        reduce_dims: Optional set of additional dimensions to aggregate over
            (mutually exclusive to ``preserve_dims``).
        preserve_dims: Optional set of dimensions to keep (all other dimensions
            are reduced); By default all dimensions except ``spatial_dims``
            are preserved (mutually exclusive to ``reduce_dims``).
        threshold_operator: The threshold operator used to generate the binary
            event field. E.g. ``np.greater``. Note: this may depend on the backend
            ``compute method`` and not all operators may be supported. Generally,
            ``np.greater``, ``np.less`` and their counterparts. Defaults to "np.greater".
        compute_method: currently only supports :py:obj:`FssComputeMethod.NUMPY`
            see: :py:class:`scores.fast.fss.typing.FssComputeMethod`
        dask: See ``xarray.apply_ufunc`` for options

    Returns:
        An ``xarray.DataArray`` containing the FSS computed over the ``spatial_dims``

        The resultant array will have the score grouped against the remaining
        dimensions, unless ``reduce_dims``/``preserve_dims`` are specified; in which
        case, they will be aggregated over the specified dimensions accordingly.

        For an exact usage please refer to ``FSS.ipynb`` in the tutorials.

    Raises:
        DimensionError: Various errors are thrown if the input dimensions
            do not conform. e.g. if the window size is larger than the input
            arrays, or if the the spatial dimensions in the args are missing in
            the input arrays.

        DimensionWarning: If ``spatial_dims`` are attempting to be preserved e.g. in ``preserve_dims``

    References:
        1. Roberts, N. M., and H. W. Lean, 2008: Scale-Selective Verification of Rainfall
           Accumulations from High-Resolution Forecasts of Convective Events. Monthly Weather
           Review, 136, 78–97, https://doi.org/10.1175/2007mwr2123.1.
        2. Mittermaier, M. P., 2021: A “Meta” Analysis of the Fractions Skill Score: The Limiting
           Case and Implications for Aggregation. Monthly Weather Review, 149, 3491–3504,
           https://doi.org/10.1175/mwr-d-18-0106.1.
    """
    np_thrsh_op = _make_numpy_threshold_operator(threshold_operator)

    def _spatial_dims_exist(_dims):
        s_spatial_dims = set(spatial_dims)
        s_dims = set(_dims)
        return (
            len(s_spatial_dims - s_dims) == 0
            and len(spatial_dims) == 2  # all spatial dims present  # number of spatial dims = 2
        )

    if not _spatial_dims_exist(fcst.dims):
        raise DimensionError(f"missing spatial dims {spatial_dims} in fcst")
    if not _spatial_dims_exist(obs.dims):
        raise DimensionError(f"missing spatial dims {spatial_dims} in obs")
    for s_dim in spatial_dims:
        if obs[s_dim].shape != fcst[s_dim].shape:
            raise DimensionError(
                f"The spatial extent of fcst and obs do not match for the following spatial dimensions: {spatial_dims}."
            )

    # wrapper defined for convenience, since it's too long for a lambda.
    def _fss_wrapper(da_fcst: xr.DataArray, da_obs: xr.DataArray) -> FssDecomposed:
        fss_backend = get_compute_backend(compute_method)
        fb_obj = fss_backend(
            da_fcst,
            da_obs,
            event_threshold=event_threshold,
            window_size=window_size,
            zero_padding=zero_padding,
            threshold_operator=np_thrsh_op,
        )
        return fb_obj.compute_fss_decomposed()

    # apply ufunc to get the decomposed fractions skill score aggregated over
    # the 2D spatial dims.
    da_fss = xr.apply_ufunc(
        _fss_wrapper,
        fcst,
        obs,
        input_core_dims=[list(spatial_dims), list(spatial_dims)],
        vectorize=True,
        dask=dask,  # pragma: no cover
    )

    # gather additional dimensions to aggregate over.
    dims = gather_dimensions(
        fcst.dims,
        obs.dims,
        reduce_dims=reduce_dims,
        preserve_dims=preserve_dims,
    )

    if not (spatial_dims[0] in dims and spatial_dims[1] in dims):
        # at least one spatial dim is trying to be preserved.
        warnings.warn(
            f"At least one of the provided spatial dims = {spatial_dims} is attempting "
            "to be preserved. Please make sure `preserve_dims` and `reduce_dims` "
            "are configured to not preserve `spatial_dims`, in order to suppress this "
            "warning.",
            DimensionWarning,
        )

    # trim any spatial dimensions as they shouldn't appear.
    dims = set(dims) - set(spatial_dims)

    # apply ufunc again but this time to compute the fss, reducing
    # any non-spatial dimensions.
    da_fss = xr.apply_ufunc(
        _aggregate_fss_decomposed,
        da_fss,
        input_core_dims=[list(dims)],
        vectorize=True,
        dask=dask,  # pragma: no cover
    )

    return da_fss


def fss_2d_binary(  # pylint: disable=too-many-locals,too-many-arguments
    fcst: xr.DataArray,
    obs: xr.DataArray,
    *,  # Force keywords arguments to be keyword-only
    window_size: Tuple[int, int],
    spatial_dims: Tuple[str, str],
    zero_padding: bool = False,
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    compute_method: FssComputeMethod = FssComputeMethod.NUMPY,
    check_boolean: bool = True,
    dask: str = "forbidden",  # see: `xarray.apply_ufunc` for options
) -> xr.DataArray:
    """
    Computes the Fractions Skill Score (FSS) using a binary field.

    Takes in a binary (True or False) field of fcst and obs events. For example
    the output of a binary threshold applied to a continuous field or an event
    operator.

    Optionally the user can set ``check_boolean`` to ``True`` to check that the
    field is boolean before any computations. Note: this asserts that the
    underlying fields are of type ``np.bool_`` but does not coerce them. If
    the user is confident that the input field is binary (but not necessarily
    boolean), they may set this flag to ``False``, to allow for more flexible binary
    configurations such as 0 and 1; this should give the same results.

    Uses ``fss_2d`` from the ``scores.spatial`` module to perform fss
    computation, but without the threshold operation.

    .. seealso::
        :py:func:`scores.spatial.fss_2d` for more details and the continuous
        version. As well as detailed argument definitions.
    """

    if check_boolean and not (fcst.dtype == np.bool_ and obs.dtype == np.bool_):
        raise FieldTypeError("Input field is not boolean")

    # Note: this s a dummy value that will be discarded by the
    # `left_identity_operator` mainly used to circumvent passing in "None" to
    # threshold operators which can have undefined behaviour.
    _phantom_event_threshold = -999

    return fss_2d(  # pylint: disable=too-many-locals,too-many-arguments
        fcst,
        obs,
        event_threshold=_phantom_event_threshold,
        window_size=window_size,
        spatial_dims=spatial_dims,
        zero_padding=zero_padding,
        reduce_dims=reduce_dims,
        preserve_dims=preserve_dims,
        threshold_operator=left_identity_operator,
        compute_method=compute_method,
        dask=dask,  # see: `xarray.apply_ufunc` for options
    )


def fss_2d_single_field(
    fcst: npt.NDArray[float],
    obs: npt.NDArray[float],
    *,
    event_threshold: float,
    window_size: Tuple[int, int],
    zero_padding: bool = False,
    threshold_operator: Optional[Callable] = None,
    compute_method: FssComputeMethod = FssComputeMethod.NUMPY,
) -> np.float64:
    """
    Calculates the Fractions Skill Score (FSS) :sup:`[1]` for a given forecast and observed
    2-D field.

    The FSS is computed by counting the squared sum of forecast and observation
    events in a given window size. This is repeated over all possible window
    positions that can fit in the input arrays. A common method to do this is
    via a sliding window.

    While the various compute backends may have their own implementations; most
    of them use some form of prefix sum algorithm to compute this quickly.

    For 2-D fields this data structure is known as the "Summed Area
    Table" :sup:`[2]`.

    Once the squared sums are computed, the final FSS value can be derived by
    accumulating the squared sums :sup:`[3]`.


    The caller is responsible for making sure the input fields are in the 2-D
    spatial domain. (Although it should work for any ``np.array`` as long as it's
    2D, and ``window_size`` is appropriately sized.)

    Args:
        fcst: An array of forecasts
        obs: An array of observations (same spatial shape as ``fcst``)
        event_threshold: A scalar to compare ``fcst`` and ``obs`` fields to generate a
            binary "event" field.
        window_size: A pair of positive integers ``height, width)`` of the sliding
            window; the window dimensions must be greater than 0 and fit within
            the shape of ``obs`` and ``fcst``.
        threshold_operator: The threshold operator used to generate the binary
            event field. E.g. ``np.greater``. Note: this may depend on the backend
            ``compute method`` and not all operators may be supported. Generally,
            ``np.greater``, ``np.less`` and their counterparts. Defaults to "np.greater".
        compute_method: currently only supports ``FssComputeMethod.NUMPY``
            see: :py:class:``FssComputeMethod``

    Returns:
        A float representing the accumulated FSS.

    Raises:
        DimensionError: Various errors are thrown if the input dimensions
            do not conform. e.g. if the window size is larger than the input
            arrays, or if the the spatial dimensions of the input arrays do
            not match.

    References:
        1. Roberts, N. M., and H. W. Lean, 2008: Scale-Selective Verification of Rainfall
           Accumulations from High-Resolution Forecasts of Convective Events. Monthly Weather Review
           136, 78–97, https://doi.org/10.1175/2007mwr2123.1.
        2. https://en.wikipedia.org/wiki/Summed-area_table
        3. FAGGIAN, N., B. ROUX, P. STEINLE, and B. EBERT, 2015: Fast calculation of the fractions
           skill score. MAUSAM, 66, 457–466, https://doi.org/10.54302/mausam.v66i3.555.
    """
    np_thrsh_op = _make_numpy_threshold_operator(threshold_operator)

    fss_backend = get_compute_backend(compute_method)

    fb_obj = fss_backend(
        fcst,
        obs,
        window_size=window_size,
        zero_padding=zero_padding,
        event_threshold=event_threshold,
        threshold_operator=np_thrsh_op,
    )

    fss_score = fb_obj.compute_fss()

    return fss_score


def _make_numpy_threshold_operator(threshold_operator: Optional[Callable]) -> NumpyThresholdOperator:
    thrsh_op: Callable

    if threshold_operator is None:
        thrsh_op = np.greater
    else:
        thrsh_op = threshold_operator

    return NumpyThresholdOperator(thrsh_op)


def _aggregate_fss_decomposed(fss_d: FssDecomposed) -> np.float64:
    """
    Aggregates the results of decomposed fss scores.
    """
    # can't do ufuncs over custom void types currently...
    fcst_sum = 0.0
    obs_sum = 0.0
    diff_sum = 0.0

    if isinstance(fss_d, np.ndarray):
        l = fss_d.size

        if l < 1:
            return 0.0

        with np.nditer(fss_d) as it:
            for elem in it:
                (fcst_, obs_, diff_) = elem.item()
                fcst_sum += fcst_ / l
                obs_sum += obs_ / l
                diff_sum += diff_ / l
    else:
        (fcst_sum, obs_sum, diff_sum) = fss_d

    fss = 0.0
    denom = obs_sum + fcst_sum

    if denom > 0.0:
        fss = 1.0 - diff_sum / denom

    fss_clamped = max(min(fss, 1.0), 0.0)

    return fss_clamped



"""Tools for discretising data for verification"""

import operator
from collections.abc import Iterable
from typing import Optional, Union

import numpy as np
import xarray as xr

from scores.typing import FlexibleDimensionTypes, XarrayLike
from scores.utils import gather_dimensions

INEQUALITY_MODES = {
    ">=": (operator.ge, -1),
    ">": (operator.gt, 1),
    "<=": (operator.le, 1),
    "<": (operator.lt, -1),
}
# '==' does not map to `operater.eq` and '!=' does not map to operator.ne on purpose.
# This is because we wish to test within a tolerance.
EQUALITY_MODES = {"==": (operator.le), "!=": (operator.gt)}


def comparative_discretise(
    data: XarrayLike, comparison: Union[xr.DataArray, float, int], mode: str, *, abs_tolerance: Optional[float] = None
) -> XarrayLike:
    """
    Converts the values of `data` to 0 or 1 based on how they relate to the specified
    values in `comparison` via the `mode` operator.

    Args:
        data: The data to convert to discrete values.
        comparison: The values to which to compare `data`.
        mode: Specifies the required relation of `data` to `thresholds`
            for a value to fall in the 'event' category (i.e. assigned to 1).
            Allowed modes are:
            - '>=' values in `data` greater than or equal to the
            corresponding threshold are assigned as 1.
            - '>' values in `data` greater than the corresponding threshold
            are assigned as 1.
            - '<=' values in `data` less than or equal to the corresponding
            threshold are assigned as 1.
            - '<' values in `data` less than the corresponding threshold
            are assigned as 1.
            - '==' values in `data` equal to the corresponding threshold
            are assigned as 1
            - '!=' values in `data` not equal to the corresponding threshold
            are assigned as 1.
        abs_tolerance: If supplied, values in data that are
            within abs_tolerance of a threshold are considered to be equal to
            that threshold. This is generally used to correct for floating
            point rounding, e.g. we may want to consider 1.0000000000000002 as
            equal to 1.
    Returns:
        An xarray data object of the same type as `data`. The dimensions of the
        output are the union of all dimensions in `data` and `comparison`. The
        values of the output are either 0 or 1 or NaN, depending on the truth
        of the operation `data <mode> comparison`.
    Raises:
       ValueError: if abs_tolerance is not a non-negative float.
       ValueError: if `mode` is not valid.
       TypeError: if `comparison` is not a float, int or xarray.DataArray.
    """

    # sanitise abs_tolerance
    if abs_tolerance is None:
        abs_tolerance = 0
    elif abs_tolerance < 0:
        raise ValueError(f"value {abs_tolerance} of abs_tolerance is invalid, it must be a non-negative float")

    if isinstance(comparison, (float, int)):
        comparison = xr.DataArray(comparison)
    elif not isinstance(comparison, xr.DataArray):
        raise TypeError("comparison must be a float, int or xarray.DataArray")

    # mask to preserve NaN in data and comparison
    notnull_mask = data.notnull() * comparison.notnull()

    # do the discretisation
    if mode in INEQUALITY_MODES:
        operator_func, factor = INEQUALITY_MODES[mode]
        discrete_data = operator_func(data, comparison + (abs_tolerance * factor)).where(notnull_mask)
    elif mode in EQUALITY_MODES:
        operator_func = EQUALITY_MODES[mode]
        discrete_data = operator_func(abs(data - comparison), abs_tolerance).where(notnull_mask)
    else:
        raise ValueError(
            f"'{mode}' is not a valid mode. Available modes are: "
            f"{sorted(INEQUALITY_MODES) + sorted(EQUALITY_MODES)}"
        )
    discrete_data.attrs["discretisation_tolerance"] = abs_tolerance
    discrete_data.attrs["discretisation_mode"] = mode

    return discrete_data  # type: ignore


def binary_discretise(
    data: XarrayLike,
    thresholds: Optional[FlexibleDimensionTypes],
    mode: str,
    *,  # Force keywords arguments to be keyword-only
    abs_tolerance: Optional[float] = None,
    autosqueeze: Optional[bool] = False,
):
    """
    Converts the values of `data` to 0 or 1 for each threshold in `thresholds`
    according to the operation defined by `mode`.

    Args:
        data: The data to convert to discrete values.
        thresholds: Threshold(s) at which to convert the values of `data` to 0 or 1.
        mode: Specifies the required relation of `data` to `thresholds`
            for a value to fall in the 'event' category (i.e. assigned to 1).
            Allowed modes are:

            - '>=' values in `data` greater than or equal to the \
            corresponding threshold are assigned as 1.
            - '>' values in `data` greater than the corresponding threshold \
            are assigned as 1.
            - '<=' values in `data` less than or equal to the corresponding \
            threshold are assigned as 1.
            - '<' values in `data` less than the corresponding threshold \
            are assigned as 1.
            - '==' values in `data` equal to the corresponding threshold \
            are assigned as 1
            - '!=' values in `data` not equal to the corresponding threshold \
            are assigned as 1.

        abs_tolerance: If supplied, values in data that are
            within abs_tolerance of a threshold are considered to be equal to
            that threshold. This is generally used to correct for floating
            point rounding, e.g. we may want to consider 1.0000000000000002 as
            equal to 1

        autosqueeze: If True and only one threshold is
            supplied, then the dimension 'threshold' is squeezed out of the
            output. If `thresholds` is float-like, then this is forced to
            True, otherwise defaults to False.

    Returns:
        An xarray data object with the type and dimensions of `data`, plus an
        extra dimension 'threshold' if `autosqueeze` is False. The values of
        the output are either 0 or 1, depending on whether `data <mode> threshold`
        is True or not (although NaNs are preserved).

    Raises:
        ValueError: if 'threshold' is a dimension in `data`.
        ValueError: if "Values in `thresholds` are not monotonic increasing"
    """
    if "threshold" in data.dims:
        raise ValueError("'threshold' must not be in the supplied data object dimensions")

    # if thresholds is 0-D, convert it to a length-1 1-D array
    # but autosqueeze=True so the 'threshold' dimension is dropped
    thresholds_np = np.array(thresholds)
    if thresholds_np.ndim == 0:
        thresholds_np = np.expand_dims(thresholds_np, 0)
        autosqueeze = True

    # sanitise thresholds
    if not (thresholds_np[1:] - thresholds_np[:-1] >= 0).all():
        raise ValueError("Values in `thresholds` are not monotonic increasing")

    # make thresholds DataArray
    thresholds_da = xr.DataArray(thresholds_np, dims=["threshold"], coords={"threshold": thresholds_np})

    # do the discretisation
    discrete_data = comparative_discretise(data, thresholds_da, mode, abs_tolerance=abs_tolerance)

    # squeeze
    if autosqueeze and len(thresholds_np) == 1:
        # squeeze out the 'threshold' dimension, but keep the coordinate
        discrete_data = discrete_data.squeeze(dim="threshold")

    return discrete_data


def proportion_exceeding(
    data: XarrayLike,
    thresholds: Iterable,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
):
    """
    Calculates the proportion of `data` equal to or exceeding `thresholds`.

    Args:
        data (xarray.Dataset or xarray.DataArray): The data from which
            to calculate the proportion exceeding `thresholds`
        thresholds (iterable): The proportion of values
            equal to or exceeding these thresholds will be calculated.
        reduce_dims: Dimensions to reduce.
        preserve_dims: Dimensions to preserve.

    Returns:
        An xarray data object with the type of `data` and dimensions
        `dims` + 'threshold'. The values are the proportion of `data`
        that are greater than or equal to the corresponding threshold.

    """
    return binary_discretise_proportion(data, thresholds, ">=", reduce_dims=reduce_dims, preserve_dims=preserve_dims)


def binary_discretise_proportion(
    data: XarrayLike,
    thresholds: Iterable,
    mode: str,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    abs_tolerance: Optional[bool] = None,
    autosqueeze: bool = False,
):
    """
    Returns the proportion of `data` in each category. The categories are
    defined by the relationship of data to threshold as specified by
    the operation `mode`.

    Args:
        data: The data to convert
           into 0 and 1 according the thresholds before calculating the
           proportion.
        thresholds: The proportion of values
            equal to or exceeding these thresholds will be calculated.
        mode: Specifies the required relation of `data` to `thresholds`
            for a value to fall in the 'event' category (i.e. assigned to 1).
            Allowed modes are:

            - '>=' values in `data` greater than or equal to the
              corresponding threshold are assigned as 1.
            - '>' values in `data` greater than the corresponding threshold
              are assigned as 1.
            - '<=' values in `data` less than or equal to the corresponding
              threshold are assigned as 1.
            - '<' values in `data` less than the corresponding threshold
              are assigned as 1.
            - '==' values in `data` equal to the corresponding threshold
              are assigned as 1
            - '!=' values in `data` not equal to the corresponding threshold
              are assigned as 1.
        reduce_dims: Dimensions to reduce.
        preserve_dims: Dimensions to preserve.
        abs_tolerance: If supplied, values in data that are
            within abs_tolerance of a threshold are considered to be equal to
            that threshold. This is generally used to correct for floating
            point rounding, e.g. we may want to consider 1.0000000000000002 as
            equal to 1.
        autosqueeze: If True and only one threshold is
            supplied, then the dimension 'threshold' is squeezed out of the
            output. If `thresholds` is float-like, then this is forced to
            True, otherwise defaults to False.

    Returns:
        An xarray data object with the type of `data`, dimension `dims` +
        'threshold'. The values of the output are the proportion of `data` that
        satisfy the relationship to `thresholds` as specified by `mode`.

    Examples:

        >>> data = xr.DataArray([0, 0.5, 0.5, 1])

        >>> _binary_discretise_proportion(data, [0, 0.5, 1], '==')
        <xarray.DataArray (threshold: 3)>
        array([ 0.25,  0.5 ,  0.25])
        Coordinates:
          * threshold  (threshold) float64 0.0 0.5 1.0
        Attributes:
            discretisation_tolerance: 0
            discretisation_mode: ==

        >>> _binary_discretise_proportion(data, [0, 0.5, 1], '>=')
        <xarray.DataArray (threshold: 3)>
        array([ 1.  ,  0.75,  0.25])
        Coordinates:
          * threshold  (threshold) float64 0.0 0.5 1.0
        Attributes:
            discretisation_tolerance: 0
            discretisation_mode: >=

    See also:
        `scores.processing.binary_discretise`

    """
    # values are 1 when (data {mode} threshold), and 0 when ~(data {mode} threshold).
    discrete_data = binary_discretise(data, thresholds, mode, abs_tolerance=abs_tolerance, autosqueeze=autosqueeze)

    # The proportion in each category
    dims = gather_dimensions(data.dims, data.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims)
    proportion = discrete_data.mean(dim=dims)

    # attach attributes
    proportion.attrs = discrete_data.attrs

    return proportion



"""
An implementation of isotonic regression using the pool adjacent violators (PAV)
algorithm. In the context of forecast verification, the regression explanatory variable
is the forecast and the response variable is the observation. Forecasts and observations
are assumed to be real-valued quantities.

Isotonic regression exposes conditional forecast biases. Confidence bands for the
regression fit can also be returned. This implementation includes option to specify what
functional the forecast is targeting, so that appropriate regression can be performed on
different types of real-valued forecasts, such as those which are a quantile of the
predictive distribution.
"""

from functools import partial
from typing import Callable, Literal, Optional, Tuple, Union

import numpy as np
import xarray as xr
from scipy import interpolate
from sklearn.isotonic import IsotonicRegression


def isotonic_fit(  # pylint: disable=too-many-locals, too-many-arguments
    fcst: Union[np.ndarray, xr.DataArray],
    obs: Union[np.ndarray, xr.DataArray],
    *,  # Force keywords arguments to be keyword-only
    weight: Optional[Union[np.ndarray, xr.DataArray]] = None,
    functional: Optional[Literal["mean", "quantile"]] = "mean",
    bootstraps: Optional[int] = None,
    quantile_level: Optional[float] = None,
    solver: Optional[Union[Callable[[np.ndarray, np.ndarray], float], Callable[[np.ndarray], float]]] = None,
    confidence_level: Optional[float] = 0.9,
    min_non_nan: Optional[int] = 1,
    report_bootstrap_results: Optional[bool] = False,
) -> dict:
    """
    Calculates an isotonic regression fit to observations given forecasts as the
    explanatory values. Optionally returns confidence bands on the fit via bootstrapping.
    Forecasts and observations are scalar-valued (i.e, integers, floats or NaN).

    If forecasts target the mean or a quantile functional, then the user can select that
    functional. Otherwise the user needs to supply a `solver` function

        - of one variable (if no weights are supplied), or
        - of two variables (if weights are supplied).

    The solver function is used to find an optimal regression fit in each block, subject
    to a non-decreasing monotonicity constraint, using the pool adjacent violators (PAV)
    algorithm.

    An example of a solver function of one variable is obs -> np.mean(obs)
    An example of a solver function of two variables is (obs, weight) -> np.average(obs, weights=weight).

    Ties (where one forecast value has more than one observation value) are handled as follows.
    Forecasts are sorted in ascending order, and observations sorted to match. Where ties occur,
    observations are sorted in descending order within each subset of ties. Isotonic regression
    is then performed on this sorted observation sequence.

    This implementation uses sklearn.isotonic.IsotonicRegression when `functional="mean"`.

    Args:
        fcst: np.array or xr.DataArray of forecast (or explanatory) values. Values must
            be float, integer or NaN.
        obs: np.array or xr.DataArray of corresponding observation (or response) values.
            Must be the same shape as `fcst`. Values must be float, integer or NaN.
        weight: positive weights to apply to each forecast-observation pair. Must be the
            same shape as `fcst`, or `None` (which is equivalent to applying equal weights).
        functional: Functional that the forecasts are targeting. Either "mean" or
            "quantile" or `None`. If `None` then `solver` must be supplied. The current
            implementation for "quantile" does not accept weights. If weighted quantile
            regression is desired then the user should should supply an appropriate solver.
        bootstraps: the number of bootstrap samples to perform for calculating the
            regression confidence band. Set to `None` if a confidence band is not required.
        quantile_level: the level of the quantile functional if `functional='quantile'`.
            Must be strictly between 0 and 1.
        solver: function that accepts 1D numpy array of observations and returns
            a float. Function values give the regression fit for each block as determined
            by the PAV algorithm. If weights are supplied, the function must have a second
            argument that accepts 1D numpy array of weights of same length as the observations.
        confidence_level: Confidence level of the confidence band, strictly between 0 and 1.
            For example, a confidence level of 0.5 will calculate the confidence band by
            computing the 25th and 75th percentiles of the bootstrapped samples.
        min_non_nan: The minimum number of non-NaN bootstrapped values to calculate
            confidence bands at a particular forecast value.
        report_bootstrap_results: This specifies to whether keep the bootstrapped
            regression values in return dictionary. Default is False to keep the output
            result small.

    Returns:
        Dictionary with the following keys:

        - "unique_fcst_sorted": 1D numpy array of remaining forecast values sorted in
            ascending order, after any NaNs from `fcst`, `obs` and `weight` are removed,
            and only unique values are kept to keep the output size reasonably small.
        - "fcst_counts": 1D numpy array of forecast counts for unique values of forecast sorted
        - "regression_values": 1D numpy array of regression values corresponding to
            "unique_fcst_sorted" values.
        - "regression_func": function that returns the regression fit based on linear
            interpolation of ("fcst_sorted", "regression_values"), for any supplied
            argument (1D numpy array) of potential forecast values.
        - "bootstrap_results": in the case of `report_bootstrap_results=True`, 2D numpy
            array of bootstrapped regression values is included in return dictionary.
            Each row gives the interpolated regression values from a particular bootstrapped
            sample, evaluated at "fcst_sorted" values. If `m` is the number of bootstrap
            samples and `n = len(fcst_sorted)` then it is has shape `(m, n)`. We emphasise
            that this array corresponds to `fcst_sorted` not `unique_fcst_sorted`.
        - "confidence_band_lower_values": values of lower confidence band threshold, evaluated
            at "unique_fcst_sorted" values.
        - "confidence_band_upper_values": values of upper confidence band threshold, evaluated
            at "unique_fcst_sorted" values.
        - "confidence_band_lower_func": function that returns regression fit based on linear
            interpolation of ("fcst_sorted", "confidence_band_lower_values"), given any
            argument (1D numpy array) of potential forecast values.
        - "confidence_band_upper_func": function that returns regression fit based on linear
            interpolation of ("fcst_sorted", "confidence_band_upper_values"), given any
            argument (1D numpy array) of potential forecast values.
        - "confidence_band_levels": tuple giving the quantile levels used to calculate the
            confidence band.

    Raises:
        ValueError: if `fcst` and `obs` are np.arrays and don't have the same shape.
        ValueError: if `fcst` and `weight` are np.arrays and don't have the same shape.
        ValueError: if `fcst` and `obs` are xarray.DataArrays and don't have the same dimensions.
        ValueError: if `fcst` and `weight` are xarray.DataArrays and don't have the same dimensions.
        ValueError: if any entries of `fcst`, `obs` or `weight` are not an integer, float or NaN.
        ValueError: if there are no `fcst` and `obs` pairs after NaNs are removed.
        ValueError: if `functional` is not one of "mean", "quantile" or `None`.
        ValueError: if `quantile_level` is not strictly between 0 and 1 when `functional="quantile"`.
        ValueError: if `weight` is not `None` when `functional="quantile"`.
        ValueError: if `functional` and `solver` are both `None`.
        ValueError: if not exactly one of `functional` and `solver` is `None`.
        ValueError: if `bootstraps` is not a positive integer.
        ValueError: if `confidence_level` is not strictly between 0 and 1.

    Note: This function only keeps the unique values of `fcst_sorted` to keep the volume of
        the return dictionary small. The forecast counts is also included, so users can it to
        create forecast histogram (usually displayed in the reliability diagrams).

    References
        - de Leeuw, Hornik and Mair. "Isotone Optimization in R: Pool-Adjacent-Violators Algorithm (PAVA)
          and Active Set Methods", Journal of Statistical Software, 2009.
        - Dimitriadis, Gneiting and Jordan. "Stable reliability diagrams for probabilistic classifiers",
          PNAS, Vol. 118 No. 8, 2020. Available at https://www.pnas.org/doi/10.1073/pnas.2016191118
        - Jordan, Mühlemann, and Ziegel. "Optimal solutions to the isotonic regression problem",
          2020 (version 2), available on arxiv at https://arxiv.org/abs/1904.04761
    """

    if isinstance(fcst, xr.DataArray):
        fcst, obs, weight = _xr_to_np(fcst, obs, weight)  # type: ignore
    # now fcst, obs and weight (unless None) are np.arrays

    _iso_arg_checks(
        fcst,  # type: ignore
        obs,  # type: ignore
        weight=weight,  # type: ignore
        functional=functional,
        quantile_level=quantile_level,
        solver=solver,
        bootstraps=bootstraps,
        confidence_level=confidence_level,
    )
    fcst_tidied, obs_tidied, weight_tidied = _tidy_ir_inputs(fcst, obs, weight=weight)  # type: ignore
    y_out = _do_ir(
        fcst_tidied,
        obs_tidied,
        weight=weight_tidied,
        functional=functional,
        quantile_level=quantile_level,
        solver=solver,
    )

    # calculate the fitting function
    ir_func = interpolate.interp1d(fcst_tidied, y_out, bounds_error=False)

    if bootstraps is not None:
        boot_results = _bootstrap_ir(
            fcst=fcst_tidied,
            obs=obs_tidied,
            weight=weight_tidied,
            functional=functional,
            quantile_level=quantile_level,
            solver=solver,
            bootstraps=bootstraps,
        )

        lower_pts, upper_pts = _confidence_band(boot_results, confidence_level, min_non_nan)  # type: ignore

        lower_func = interpolate.interp1d(fcst_tidied, lower_pts, bounds_error=False)
        upper_func = interpolate.interp1d(fcst_tidied, upper_pts, bounds_error=False)

        confband_levels = ((1 - confidence_level) / 2, 1 - (1 - confidence_level) / 2)  # type: ignore

    else:
        boot_results = lower_pts = upper_pts = None  # type: ignore
        lower_func = upper_func = partial(np.full_like, fill_value=np.nan)
        confband_levels = (None, None)  # type: ignore
    # To reduce the size of output dictionary, we only keep the unique values of
    # forecasts and accordingly regression and confidence band values calculate by using
    # unique forecast values. We also calculate forecast counts that can be used to create
    # the forecast histogram.
    unique_fcst_sorted, fcst_counts = np.unique(fcst_tidied, return_counts=True)
    regression_values = ir_func(unique_fcst_sorted)
    if bootstraps:
        lower_pts = lower_func(unique_fcst_sorted)
        upper_pts = upper_func(unique_fcst_sorted)
    results = {
        "fcst_sorted": unique_fcst_sorted,
        "fcst_counts": fcst_counts,
        "regression_values": regression_values,
        "regression_func": ir_func,
        "confidence_band_lower_values": lower_pts,
        "confidence_band_upper_values": upper_pts,
        "confidence_band_lower_func": lower_func,
        "confidence_band_upper_func": upper_func,
        "confidence_band_levels": confband_levels,
    }
    if report_bootstrap_results:
        results["bootstrap_results"] = boot_results

    return results


def _xr_to_np(
    fcst: xr.DataArray, obs: xr.DataArray, weight: Optional[xr.DataArray]
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Conducts basic dimension checks to `isotonic_fit` inputs if they are xr.DataArray.
    Then converts xr.DataArray objects into corresponding numpy arrays, with
    corresponding datapoints.

    Args:
        fcst: DataArray of forecast values
        obs: DataArray of observed values
        weight: weights to apply to each forecast-observation pair. Must be the same
            shape as `fcst`, or `None`.
    Returns:
        numpy arrays of `fcst`, `obs` and `weight`.

    Raises:
        ValueError: if `fcst` and `obs` do not have the same dimension.
        ValueError: if `fcst` and `weight` do not have the same dimension.
    """
    fcst_dims = fcst.dims

    if set(fcst_dims) != set(obs.dims):
        raise ValueError("`fcst` and `obs` must have same dimensions.")

    if weight is not None:
        if set(fcst_dims) != set(weight.dims):
            raise ValueError("`fcst` and `weight` must have same dimensions.")
        merged_ds = xr.merge([fcst.rename("fcst"), obs.rename("obs"), weight.rename("weight")])
        weight = merged_ds["weight"].transpose(*fcst_dims).values  # type: ignore
    else:
        merged_ds = xr.merge([fcst.rename("fcst"), obs.rename("obs")])

    fcst = merged_ds["fcst"].transpose(*fcst_dims).values  # type: ignore
    obs = merged_ds["obs"].transpose(*fcst_dims).values  # type: ignore

    return fcst, obs, weight  # type: ignore


def _iso_arg_checks(  # pylint: disable=too-many-arguments, too-many-branches
    fcst: np.ndarray,
    obs: np.ndarray,
    *,  # Force keywords arguments to be keyword-only
    weight: Optional[Union[np.ndarray, xr.DataArray]] = None,
    functional: Optional[str] = None,
    quantile_level: Optional[float] = None,
    solver: Optional[Union[Callable[[np.ndarray, np.ndarray], float], Callable[[np.ndarray], float]]] = None,
    bootstraps: Optional[int] = None,
    confidence_level: Optional[float] = None,
) -> None:
    """Raises ValueError if isotonic_fit arguments are invalid."""
    if fcst.shape != obs.shape:
        raise ValueError("`fcst` and `obs` must have same shape.")

    if not (np.issubdtype(fcst.dtype, np.integer) or np.issubdtype(fcst.dtype, np.floating)):
        raise ValueError("`fcst` must be an array of floats or integers.")

    if not (np.issubdtype(obs.dtype, np.integer) or np.issubdtype(obs.dtype, np.floating)):
        raise ValueError("`obs` must be an array of floats or integers.")

    if weight is not None:
        if fcst.shape != weight.shape:
            raise ValueError("`fcst` and `weight` must have same shape.")

        if not (np.issubdtype(weight.dtype, np.integer) or np.issubdtype(weight.dtype, np.floating)):
            raise ValueError("`weight` must be an array of floats or integers, or else `None`.")

        if np.any(weight <= 0):
            raise ValueError("`weight` values must be either positive or NaN.")

    if functional not in ["mean", "quantile", None]:
        raise ValueError("`functional` must be one of 'mean', 'quantile' or `None`.")

    if functional == "quantile" and not 0 < quantile_level < 1:  # type: ignore
        raise ValueError("`quantile_level` must be strictly between 0 and 1.")

    if functional == "quantile" and weight is not None:
        msg = "Weighted quantile isotonic regression has not been implemented. "
        msg += "Either (i) set `weight=None` or (ii) set `functional=None` and supply an appropriate "
        msg += "quantile `solver` function with two arguments."
        raise ValueError(msg)

    if functional is None and solver is None:
        raise ValueError("`functional` and `solver` cannot both be `None`.")

    if None not in (functional, solver):
        raise ValueError("One of `functional` or `solver` must be `None`.")

    if bootstraps is not None:
        if not isinstance(bootstraps, int) or bootstraps < 1:
            raise ValueError("`bootstraps` must be a positive integer.")
        if not 0 < confidence_level < 1:  # type: ignore
            raise ValueError("`confidence_level` must be strictly between 0 and 1.")


def _tidy_ir_inputs(
    fcst: np.ndarray,
    obs: np.ndarray,
    *,  # Force keywords arguments to be keyword-only
    weight: Optional[np.ndarray] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Tidies array inputs to `isotonic_fit` in preparation for the regression routine.

    Arrays are flattened to 1D arrays. NaNs are jointly removed from fcst, obs, weights.
    The array data is then sorted by fcst (ascending) then by obs (descending).

    Args:
        fcst: numpy array of forecast values.
        obs: numpy array of corresponding observed values, with the same shape as `fcst`.
        weight: Either `None`, or numpy array of corresponding weights with the same
            shape as `fcst`.

    Returns:
        Tidied fcst, obs and weight arrays. If input weight is `None` then returned weight
        is `None`.

    Raises:
        ValueError: if no forecast and observation pairs remains after removing NaNs.
    """
    fcst = fcst.flatten()
    obs = obs.flatten()

    if weight is not None:
        weight = weight.flatten()

    nan_locs = np.isnan(fcst) | np.isnan(obs)

    if weight is not None:
        nan_locs = nan_locs | np.isnan(weight)

    new_fcst = fcst[~nan_locs]
    new_obs = obs[~nan_locs]

    if len(new_fcst) == 0:
        raise ValueError("No (fcst, obs) pairs remaining after NaNs removed.")

    # deal with ties in fcst: sort by fcst (ascending) then by obs (descending)
    # this ensures that ties in fcst have same regression value in PAV algorithm
    sorter = np.lexsort((-new_obs, new_fcst))
    new_fcst = new_fcst[sorter]
    new_obs = new_obs[sorter]

    if weight is None:
        new_weight = None
    else:
        new_weight = weight[~nan_locs]
        new_weight = new_weight[sorter]

    return new_fcst, new_obs, new_weight  # type: ignore


def _do_ir(  # pylint: disable=too-many-arguments
    fcst: np.ndarray,
    obs: np.ndarray,
    *,  # Force keywords arguments to be keyword-only
    weight: Optional[np.ndarray] = None,
    functional: Optional[Literal["mean", "quantile"]] = None,
    quantile_level: Optional[float] = None,
    solver: Optional[Union[Callable[[np.ndarray, np.ndarray], float], Callable[[np.ndarray], float]]] = None,
) -> np.ndarray:
    """
    Returns the isotonic regression (IR) fit for specified functional or solver,
    by passing the inputs to the appropriate IR algorithm.

    The inputs `fcst`, `obs` and `weight` are 1D numpy arrays assumed to have been
    mutually tidied via `_tidy_ir_inputs`.

    Args:
        fcst: tidied array of forecast values.
        obs: tidied array of observed values.
        weight: tidied array of weights.
        functional: either "mean", "quantile" or None.
        quantile_level: float strictly between 0 and 1 if functional="quantile",
            else None.
        solver: solver function as described in docstring of `isotonic_fit`, or `None`.
            Cannot be `None` if `functional` is `None`.

    Returns:
        1D numpy array of values fitted using isotonic regression.
    """
    if functional == "mean":
        y_out = _contiguous_mean_ir(fcst, obs, weight=weight)
    elif functional == "quantile":
        y_out = _contiguous_quantile_ir(obs, quantile_level)  # type: ignore
    else:
        y_out = _contiguous_ir(obs, solver, weight=weight)  # type: ignore

    return y_out


def _contiguous_ir(
    y: np.ndarray,
    solver: Union[Callable[[np.ndarray, np.ndarray], float], Callable[[np.ndarray], float]],
    *,  # Force keywords arguments to be keyword-only
    weight: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Given a 1D numpy array `y` of response values, returns fitted isotonic regression
    with values for each valid regression block determined by `solver`.

    This implementation uses the pool adjacent violators (PAV) algorithm, and is based on the squared loss
    implementation at sklearn._isotonic._inplace_contiguous_isotonic_regression
    https://github.com/scikit-learn/scikit-learn/blob/a6574655478a24247189236ce5f824e65ad0d369/sklearn/_isotonic.pyx
    For an animated visualisation of the algorithm, see
    https://www.r-bloggers.com/2020/05/what-is-isotonic-regression/

    Args:
        y: 1D numpy array of response (i.e., observed) values.
        solver: function which determines values of each valid regression block
            in the PAV algorithm, given response values in the block.
        weight: 1D numpy array of weights, or else `None`.
            If `weight` is not `None` then `solver` takes in two arrays as arguments
            (response values and weights). If `weight` is `None` then `solver` takes
            in a single array as argument (response values).

    """
    len_y = len(y)

    if weight is not None and len(weight) != len_y:
        raise ValueError("`y` and `weight` must have same length.")

    # y_out will be the final regression output
    y_out = y.copy()

    # target describes a list of blocks.  At any time, if [i..j] (inclusive) is
    # an active block, then target[i] := j and target[j] := i.
    target = np.arange(len_y, dtype=np.intp)

    index = 0
    while index < len_y:
        # idx_next_block is the index of beginning of the next block
        idx_next_block = target[index] + 1

        if idx_next_block == len_y:  # there are no more blocks
            break
        if (
            y_out[index] < y_out[idx_next_block]
        ):  # the regression value of the next block is greater than the active one
            index = idx_next_block  # that block becomes the active block
            continue

        while True:
            # We are within a decreasing subsequence.
            prev_y = y_out[idx_next_block]  # this is the smaller y value after the active block
            idx_next_block = target[idx_next_block] + 1  # advance index of next block
            if idx_next_block == len_y or prev_y < y_out[idx_next_block]:
                # update block indices
                target[index] = idx_next_block - 1
                target[idx_next_block - 1] = index

                # regression value for current block
                if weight is None:
                    y_out[index] = solver(y[index:idx_next_block])  # type: ignore
                else:
                    y_out[index] = solver(y[index:idx_next_block], weight[index:idx_next_block])  # type: ignore

                if index > 0:
                    # Backtrack if we can.  This makes the algorithm
                    # single-pass and ensures O(n) complexity (subject to solver complexity)
                    index = target[index - 1]
                # Otherwise, restart from the same point.
                break
    # Reconstruct the solution.
    index = 0
    while index < len_y:
        idx_next_block = target[index] + 1
        y_out[index + 1 : idx_next_block] = y_out[index]
        index = idx_next_block

    return y_out


def _contiguous_quantile_ir(y: np.ndarray, alpha: float) -> np.ndarray:
    """Performs contiguous quantile IR on tidied data y, for quantile-level alpha, with no weights."""
    return _contiguous_ir(y, partial(np.quantile, q=alpha))  # type: ignore


def _contiguous_mean_ir(
    x: np.ndarray, y: np.ndarray, *, weight: Optional[np.ndarray] = None  # Force keywords arguments to be keyword-only
) -> np.ndarray:
    """
    Performs classical (i.e. for mean functional) contiguous quantile IR on tidied data x, y.
    Uses sklearn implementation rather than supplying the mean solver function to `_contiguous_ir`,
    as it is about 4 times faster (since it is optimised for mean).
    """
    return IsotonicRegression().fit_transform(x, y, sample_weight=weight)  # type: ignore


def _bootstrap_ir(  # pylint: disable=too-many-arguments, too-many-locals
    fcst: np.ndarray,
    obs: np.ndarray,
    bootstraps: int,
    *,  # Force keywords arguments to be keyword-only
    weight: Optional[np.ndarray] = None,
    functional: Optional[Literal["mean", "quantile"]] = None,
    quantile_level: Optional[float] = None,
    solver: Optional[Union[Callable[[np.ndarray, np.ndarray], float], Callable[[np.ndarray], float]]] = None,
):
    """
    Gives the isotonic fits of bootstrapped samples.

    Args:
        fcst: 1D numpy array of forecast values.
        obs: 1D numpy array of observed values corresponding to `fcst`.
        bootstraps: the number of samples to bootstrap.
        weight: either `None`, or 1D numpy array of weights corresponding to `fcst`.
        functional: either `None` or the target functional (one of "mean" or "quantile").
        quantile_level: the quantile level if `functional="quantile"`.
        solver: the regression solver function if `functional=None`.

    Returns:
        2D numpy array with one row for each bootstrapped sample and one column for each `fcst` value.
        The [i,j] entry is the isotonic fit of the ith bootstrapped sample evaluated at the point
        `fcst[j]`. Evaluations are determined using interpolation when necessary. If the [i,j] entry is
        NaN then `fcst[j]` lay outside the range of the ith bootstrapped sample.
    """
    # initialise
    fc_values_count = len(fcst)
    result = np.full((bootstraps, fc_values_count), np.nan)

    for boostrap_sample_num in range(bootstraps):
        selection = np.random.randint(0, fc_values_count, fc_values_count)
        fcst_sample = fcst[selection]
        obs_sample = obs[selection]
        if weight is None:
            weight_sample = None
        else:
            weight_sample = weight[selection]

        fcst_sample, obs_sample, weight_sample = _tidy_ir_inputs(fcst_sample, obs_sample, weight=weight_sample)

        ir_results = _do_ir(
            fcst_sample,
            obs_sample,
            weight=weight_sample,
            functional=functional,
            quantile_level=quantile_level,
            solver=solver,
        )

        approximation_func = interpolate.interp1d(fcst_sample, ir_results, bounds_error=False)

        result[boostrap_sample_num] = approximation_func(fcst)

    return result


def _confidence_band(
    bootstrap_results: np.ndarray, confidence_level: float, min_non_nan: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Given bootstrapped results, return the lower and upper values of the corresponding confidence band
    with specified `confidence_level`.

    Args:
        bootstrap_results: matrix output from `_bootstrap_ir`.
        confidence_level: float between 0 and 1.
        min_non_nan: the minimum number of non-NaN values a column of `bootstrap_results`
            for which the confidence band will be calculated. If the minimum is not met,
            a NaN value is returned.

    Returns:
        Tuple of two 1D numpy arrays. The first array gives is the
        `(1-confidence_level)/2`-quantile of each column of `bootstrap_results`.
        The second array gives is the `1-(1-confidence_level)/2`-quantile of each column.
    """
    sig_level = 1 - confidence_level
    upper_level = 1 - sig_level / 2
    lower_level = sig_level / 2

    # confidence band values ignoring NaNs
    upper = _nanquantile(bootstrap_results, upper_level)
    lower = _nanquantile(bootstrap_results, lower_level)

    # masking values with too many NaNs in bootstrap results
    non_nan_count = np.count_nonzero(~np.isnan(bootstrap_results), axis=0)
    upper = np.where(non_nan_count >= min_non_nan, upper, np.nan)
    lower = np.where(non_nan_count >= min_non_nan, lower, np.nan)

    return lower, upper


def _nanquantile(arr: np.ndarray, quant: float) -> np.ndarray:
    """Returns same* output as np.nanquantile but faster.

    *almost equal
    See https://github.com/numpy/numpy/issues/16575, this implementation is about 100
    times faster compared to numpy 1.24.3.

    Args:
        arr: 2D numpy array with one row for each bootstrapped sample and one column for
            each `fcst` value.
        quant: Value between 0 and 1.

    Returns:
        1D numpy array of shape `arr.shape[0]` of values at each quantile.
    """
    arr = np.copy(arr)
    valid_obs_count = np.sum(np.isfinite(arr), axis=0)
    # Replace NaN with maximum (these values will be ignored)
    if not np.isnan(arr).all():
        max_val = np.nanmax(arr)
        arr[np.isnan(arr)] = max_val
    # Sort forecast values - former NaNs will move to the end
    arr = np.sort(arr, axis=0)
    result = np.zeros(shape=[arr.shape[1]])

    # Get the index of the values at the requested quantile
    desired_position = (valid_obs_count - 1) * quant
    index_floor = np.floor(desired_position).astype(np.int32)
    index_ceil = np.ceil(desired_position).astype(np.int32)
    same_index = index_floor == index_ceil

    # Linear interpolation - take the fractional part of desired position
    floor_val = arr[index_floor, np.arange(arr.shape[1])]
    floor_frac = index_ceil - desired_position
    floor_frac_val = arr[index_floor, np.arange(arr.shape[1])] * floor_frac
    ceil_frac = desired_position - index_floor
    ceil_frac_val = arr[index_ceil, np.arange(arr.shape[1])] * ceil_frac
    result = np.where(same_index, floor_val, floor_frac_val + ceil_frac_val)
    return result



"""
This foundation class provides an underpinning data structure which can be used for
contingency tables of various kinds, also known as a confusion matrix.

It allows the careful setting out of when and where forecasts closely match
observations.

The binary contingency table captures true positives (hits), true negatives (correct negatives),
false positives (false alarms) and false negatives (misses).

The process of deriving a contingency table relies on the forecast data, the observation
data, and a matching or event operator. The event operator will produce the category from the
forecast and the observed data. Specific use cases may well be supported by a case-specific
matching operator.

Scores supports complex, weighted, multi-dimensional data, including in contingency tables.

Users can supply their own event operators to the top-level module functions.
"""

# pylint: disable=too-many-lines

import operator
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import xarray as xr

import scores.utils
from scores.typing import FlexibleArrayType, FlexibleDimensionTypes

DEFAULT_PRECISION = 8


class BasicContingencyManager:  # pylint: disable=too-many-public-methods
    """
    See https://scores.readthedocs.io/en/stable/tutorials/Binary_Contingency_Scores.html for
    a detailed walkthrough showing the use of this class in practice.

    A BasicContingencyManager object provides the scoring functions which are calculated
    from a contingency table. A BasicContingencyManager can be efficiently and repeatedly
    queried for a wide variety of scores.

    A BasicContingencyManager is produced when a :py:class:`BinaryContingencyManager` is
    transformed. It is also possible to create a BasicContingencyManager from event counts or
    a contingency table, although this not a common user requirement.

    A contingency table is built only from event counts, losing the connection
    to the actual event tables in their full dimensionality. The event count data is much
    smaller than the full event tables, particularly when considering very large data sets
    like Numerical Weather Prediction (NWP) data, which could be terabytes to petabytes in
    size.

    By contrast, A :py:class:`BinaryContingencyManager` retains the full event data, which provides
    some more flexbility but may reduce efficiency.
    """

    def __init__(self, counts: dict):
        """
        Compute any arrays required and store the resulting counts.
        """
        for key, arr in counts.items():
            counts[key] = arr.compute()

        self.counts = counts
        self._make_xr_table()

    def _make_xr_table(
        self,
    ):
        """
        From a dictionary of the skill score elements, produce an xarray
        structure which corresponds.
        """
        ctable_list = []
        entry_name = []
        for i, j in self.counts.items():
            entry_name.append(i)
            ctable_list.append(j)

        xr_table = xr.concat(ctable_list, dim="contingency")
        xr_table["contingency"] = entry_name

        self.xr_table = xr_table

    def __str__(self):
        heading = "Contingency Manager (xarray view of table):"
        table = self.xr_table
        tablerepr = repr(table)
        final = "\n".join([heading, tablerepr])
        return final

    def get_counts(self) -> dict:
        """

        Returns:
            dict: A dictionary of the contingency table counts. Values are xr.DataArray instances.
            The keys are:

            - tp_count for true positive count
            - tn_count for true negative count,
            - fp_count for false positive count
            - fn_count for false negative count
            - total_count for the total number of events.

        Here is an example of what may be returned:

        .. code-block :: python

            {'tp_count': <xarray.DataArray ()> Size: 8B array(5.),
             'tn_count': <xarray.DataArray ()> Size: 8B array(11.),
             'fp_count': <xarray.DataArray ()> Size: 8B array(1.),
             'fn_count': <xarray.DataArray ()> Size: 8B array(1.),
             'total_count': <xarray.DataArray ()> Size: 8B array(18.)}

        """
        return self.counts

    def get_table(self) -> xr.DataArray:
        """
        Returns:
            xr.DataArray: The contingency table as an xarray object. Contains
            a coordinate dimension called 'contingency'. Valid coordinates
            for this dimenstion are:

            - tp_count for true positive count
            - tn_count for true negative count
            - fp_count for false positive count
            - fn_count for false negative count
            - total_count for the total number of events.

        Here is an example of what may be returned:

        .. code-block :: python

            array([ 5., 11.,  1.,  1., 18.])

            Coordinates:

                contingency
                (contingency)
                <U11
                'tp_count' ... 'total_count'

            Indexes:

                contingency
                PandasIndex

            Attributes: (0)

        """
        return self.xr_table

    def accuracy(self) -> xr.DataArray:
        """
        Identical to fraction correct.

        Accuracy calculates the proportion of forecasts which are correct.

        Returns:
            xr.DataArray: A DataArray containing the accuracy score

        .. math::
            \\text{accuracy} = \\frac{\\text{true positives} + \\text{true negatives}}{\\text{total count}}

        Notes:
            - Range: 0 to 1, where 1 indicates a perfect score.
            - "True positives" is the same at "hits".
            - "False negatives" is the same as "misses".

        References:
            https://www.cawcr.gov.au/projects/verification/#ACC
        """
        count_dictionary = self.counts
        correct_count = count_dictionary["tp_count"] + count_dictionary["tn_count"]
        ratio = correct_count / count_dictionary["total_count"]
        return ratio

    def base_rate(self) -> xr.DataArray:
        """
        The observed event frequency.

        Returns:
            xr.DataArray: An xarray object containing the base rate.

        .. math::
            \\text{base rate} = \\frac{\\text{true positives} + \\text{false negatives}}{\\text{total count}}

        Notes:
            - Range: 0 to 1, where 1 indicates the event occurred every time.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".

        References:
            Hogan, R. J. & Mason, I. B. (2011). Deterministic forecasts of binary events.
            In I. T. Jolliffe & D. B. Stephenson (Eds.), Forecast verification: A practitioner's guide in atmospheric science (2nd ed.,
            pp. 39-51). https://doi.org/10.1002/9781119960003.ch3
        """
        cd = self.counts
        br = (cd["tp_count"] + cd["fn_count"]) / cd["total_count"]
        return br

    def forecast_rate(self) -> xr.DataArray:
        """
        The forecast event frequency.

        Returns:
            xr.DataArray: An xarray object containing the forecast rate.

        .. math::
            \\text{forecast rate} = \\frac{\\text{true positives} + \\text{false positives}}{\\text{total count}}

        Notes:
            - Range: 0 to 1, where 1 indicates the event was forecast every time.
            - "True positives" is the same as "hits".
            - "False positives" is the same as "false alarms".

        References:
            Hogan, R. J. & Mason, I. B. (2011). Deterministic forecasts of binary events.
            In I. T. Jolliffe & D. B. Stephenson (Eds.), Forecast verification: A practitioner's guide in atmospheric science (2nd ed.,
            pp. 39-51). https://doi.org/10.1002/9781119960003.ch3
        """
        cd = self.counts
        br = (cd["tp_count"] + cd["fp_count"]) / cd["total_count"]
        return br

    def fraction_correct(self) -> xr.DataArray:
        """
        Identical to accuracy.

        Fraction correct calculates the proportion of forecasts which are correct.

        Returns:
            xr.DataArray: An xarray object containing the fraction correct.

        .. math::
            \\text{fraction correct} = \\frac{\\text{true positives} + \\text{true negatives}}{\\text{total count}}

        Notes:
            - Range: 0 to 1, where 1 indicates a perfect score.
            - "True positives" is the same as "hits".
            - "True negatives" is the same as "correct negatives".

        References:
            https://www.cawcr.gov.au/projects/verification/#ACC
        """
        return self.accuracy()

    def frequency_bias(self) -> xr.DataArray:
        """
        Identical to bias scores.

        How did the forecast frequency of "yes" events compare to the observed frequency of "yes" events?

        Returns:
            xr.DataAray: An xarray object containing the frequency bias

        .. math::
            \\text{frequency bias} = \\frac{\\text{true positives} + \\text{false positives}}{\\text{true positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to ∞ (infinity), where 1 indicates a perfect score.
            - "True positives" is the same as "hits".
            - "False positives" is the same as "false alarms".
            - "False negatives" is the same as "misses".

        References:
            https://www.cawcr.gov.au/projects/verification/#BIAS
        """
        # Note - bias_score calls this method
        cd = self.counts
        freq_bias = (cd["tp_count"] + cd["fp_count"]) / (cd["tp_count"] + cd["fn_count"])

        return freq_bias

    def bias_score(self) -> xr.DataArray:
        """
        Identical to frequence bias.

        How did the forecast frequency of "yes" events compare to the observed frequency of "yes" events?

        Returns:
            xr.DataArray: An xarray object containing the bias score

        .. math::
            \\text{frequency bias} = \\frac{\\text{true positives} + \\text{false positives}}{\\text{true positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to ∞ (infinity), where 1 indicates a perfect score.
            - "True positives" is the same as "hits".
            - "False positives" is the same as "false alarms".
            - "False negatives" is the same as "misses".

        References:
            https://www.cawcr.gov.au/projects/verification/#BIAS
        """
        return self.frequency_bias()

    def hit_rate(self) -> xr.DataArray:
        """
        Identical to true positive rate, probability of detection (POD), sensitivity and recall.

        Calculates the proportion of the observed events that were correctly forecast.

        Returns:
            xr.DataArray: An xarray object containing the hit rate

        .. math::
            \\text{true positives} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".

        References:
            https://www.cawcr.gov.au/projects/verification/#POD
        """
        return self.probability_of_detection()

    def probability_of_detection(self) -> xr.DataArray:
        """
        Probability of detection (POD) is identical to hit rate, true positive rate, sensitivity and recall.

        Calculates the proportion of the observed events that were correctly forecast.

        Returns:
            xr.DataArray: An xarray object containing the probability of detection

        .. math::
            \\text{hit rate} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 1.
            - "True positives" is the same as "hits"
            - "False negatives" is the same as "misses"

        References:
            https://www.cawcr.gov.au/projects/verification/#POD
        """
        # Note - hit_rate and sensitiviy call this function
        cd = self.counts
        pod = cd["tp_count"] / (cd["tp_count"] + cd["fn_count"])

        return pod

    def true_positive_rate(self) -> xr.DataArray:
        """
        Identical to hit rate, probability of detection (POD), sensitivity and recall.

        The proportion of the observed events that were correctly forecast.

        Returns:
            xr.DataArray: An xarray object containing the true positive rate

        .. math::
            \\text{true positive rate} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".

        References:
            https://www.cawcr.gov.au/projects/verification/#POD
        """
        return self.probability_of_detection()

    def false_alarm_ratio(self) -> xr.DataArray:
        """
        The false alarm ratio (FAR) calculates the fraction of the predicted "yes" events
        which did not eventuate (i.e., were false alarms).

        Returns:
            xr.DataArray: An xarray object containing the false alarm ratio

        .. math::
            \\text{false alarm ratio} = \\frac{\\text{false positives}}{\\text{true positives} + \\text{false positives}}

        Notes:
            - Range: 0 to 1. Perfect score: 0.
            - Not to be confused with the False Alarm Rate.
            - "False positives" is the same as "false alarms".
            - "True positives" is the same as "hits".

        References:
            https://www.cawcr.gov.au/projects/verification/#FAR
        """
        cd = self.counts
        far = cd["fp_count"] / (cd["tp_count"] + cd["fp_count"])

        return far

    def false_alarm_rate(self) -> xr.DataArray:
        """
        Identical to probability of false detection.

        What fraction of the non-events were incorrectly predicted?

        Returns:
            xr.DataArray: An xarray object containing the false alarm rate

        .. math::
            \\text{false alarm rate} = \\frac{\\text{false positives}}{\\text{true negatives} + \\text{false positives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 0.
            - Not to be confused with the false alarm ratio.
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            https://www.cawcr.gov.au/projects/verification/#POFD
        """
        # Note - probability of false detection calls this function
        cd = self.counts
        far = cd["fp_count"] / (cd["tn_count"] + cd["fp_count"])

        return far

    def probability_of_false_detection(self) -> xr.DataArray:
        """
        Probability of false detection (POFD) is identical to the false alarm rate.

        What fraction of the non-events were incorrectly predicted?

        Returns:
            xr.DataArray: An xarray object containing the probability of false detection

        .. math::
            \\text{probability of false detection} = \\frac{\\text{false positives}}{\\text{true negatives} +
            \\text{false positives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 0.
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            https://www.cawcr.gov.au/projects/verification/#POFD
        """

        return self.false_alarm_rate()

    def success_ratio(self) -> xr.DataArray:
        """
        Identical to precision.

        What proportion of the forecast events actually eventuated?

        Returns:
            xr.DataArray: An xarray object containing the success ratio

        .. math::
            \\text{success ratio} = \\frac{\\text{true positives}}{\\text{true positives} +
            \\text{false positives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False positives" is the same as "misses".

        References:
            https://www.cawcr.gov.au/projects/verification/#SR
        """
        cd = self.counts
        sr = cd["tp_count"] / (cd["tp_count"] + cd["fp_count"])

        return sr

    def threat_score(self) -> xr.DataArray:
        """
        Identical to critical success index.

        Returns:
            xr.DataArray: An xarray object containing the threat score

        .. math::
            \\text{threat score} = \\frac{\\text{true positives}}{\\text{true positives} +
            \\text{false positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to 1, 0 indicates no skill. Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            https://www.cawcr.gov.au/projects/verification/#CSI
        """
        # Note - critical success index just calls this method

        cd = self.counts
        ts = cd["tp_count"] / (cd["tp_count"] + cd["fp_count"] + cd["fn_count"])
        return ts

    def critical_success_index(self) -> xr.DataArray:
        """
        Identical to threat score.

        Returns:
            xr.DataArray: An xarray object containing the critical success index

        .. math::
            \\text{threat score} = \\frac{\\text{true positives}}{\\text{true positives} +
            \\text{false positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to 1, 0 indicates no skill. Perfect score: 1.
            - Often known as CSI.
            - "True positives" is the same as "hits"
            - "False positives" is the same as "false alarms"
            - "True negatives" is the same as "correct negatives"

        References:
            https://www.cawcr.gov.au/projects/verification/#CSI
        """
        return self.threat_score()

    def peirce_skill_score(self) -> xr.DataArray:
        """
        Identical to Hanssen and Kuipers discriminant and the true skill statistic.

        How well did the forecast separate the "yes" events from the "no" events?

        Returns:
            xr.DataArray: An xarray object containing the Peirce Skill Score

        .. math::
            \\text{Peirce skill score} = \\frac{\\text{true positives}}{\\text{true positives} + 
            \\text{false negatives}} - \\frac{\\text{false positives}}{\\text{false positives} + 
            \\text{true negatives}}

        Notes:
            - Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            - https://www.cawcr.gov.au/projects/verification/#HK
            - Peirce, C.S., 1884. The numerical measure of the success of predictions. \
              Science, ns-4(93), pp.453-454. https://doi.org/10.1126/science.ns-4.93.453.b
        """
        cd = self.counts
        component_a = cd["tp_count"] / (cd["tp_count"] + cd["fn_count"])
        component_b = cd["fp_count"] / (cd["fp_count"] + cd["tn_count"])
        skill_score = component_a - component_b
        return skill_score

    def true_skill_statistic(self) -> xr.DataArray:
        """
        Identical to Peirce's skill score and to Hanssen and Kuipers discriminant.

        How well did the forecast separate the "yes" events from the "no" events?

        Returns:
            xr.DataArray: An xarray object containing the true skill statistic

        .. math::
            \\text{true skill statistic} = \\frac{\\text{true positives}}{\\text{true positives} +
            \\text{false negatives}} - \\frac{\\text{false positives}}{\\text{false positives} +
            \\text{true negatives}}

        Notes:
            - Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            https://www.cawcr.gov.au/projects/verification/#HK
        """
        return self.peirce_skill_score()

    def hanssen_and_kuipers_discriminant(self) -> xr.DataArray:
        """
        Identical to Peirce's skill score and to the true skill statistic.

        How well did the forecast separate the "yes" events from the "no" events?

        Returns:
            xr.DataArray: An xarray object containing Hanssen and Kuipers' Discriminant

        .. math::
            \\text{HK} = \\frac{\\text{true positives}}{\\text{true positives} +
            \\text{false negatives}} - \\frac{\\text{false positives}}{\\text{false positives} +
            \\text{true negatives}}

        where :math:`\\text{HK}` is Hansen and Kuipers Discriminant

        Notes:
            - Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            https://www.cawcr.gov.au/projects/verification/#HK
        """
        return self.peirce_skill_score()

    def sensitivity(self) -> xr.DataArray:
        """
        Identical to hit rate, probability of detection (POD), true positive rate and recall.

        Calculates the proportion of the observed events that were correctly forecast.

        Returns:
            xr.DataArray: An xarray object containing the probability of detection

        .. math::
            \\text{sensitivity} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".

        References:
            - https://www.cawcr.gov.au/projects/verification/#POD
            - https://en.wikipedia.org/wiki/Sensitivity_and_specificity

        """
        return self.probability_of_detection()

    def specificity(self) -> xr.DataArray:
        """
        Identical to true negative rate.

        The probability that an observed non-event will be correctly predicted.

        Returns:
            xr.DataArray: An xarray object containing the true negative rate (specificity).

        .. math::
            \\text{specificity} = \\frac{\\text{true negatives}}{\\text{true negatives} + \\text{false positives}}

        Notes:
            - "True negatives" is the same as "correct negatives".
            - "False positives" is the same as "false alarms".

        Reference:
            https://en.wikipedia.org/wiki/Sensitivity_and_specificity
        """
        cd = self.counts
        s = cd["tn_count"] / (cd["tn_count"] + cd["fp_count"])
        return s

    def true_negative_rate(self) -> xr.DataArray:
        """
        Identical to specificity.

        The probability that an observed non-event will be correctly predicted.

        Returns:
            xr.DataArray: An xarray object containing the true negative rate.

        .. math::
            \\text{true negative rate} = \\frac{\\text{true negatives}}{\\text{true negatives} + \\text{false positives}}

        Notes:
            - "True negatives" is the same as "correct negatives".
            - "False positives" is the same as "false alarms".

        Reference:
            https://en.wikipedia.org/wiki/Sensitivity_and_specificity
        """
        return self.specificity()

    def recall(self) -> xr.DataArray:
        """
        Identical to hit rate, probability of detection (POD), true positive rate and sensitivity.

        Calculates the proportion of the observed events that were correctly forecast.

        Returns:
            xr.DataArray: An xarray object containing the probability of detection

        .. math::
            \\text{recall} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false negatives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".

        References:
            - https://www.cawcr.gov.au/projects/verification/#POD
            - https://en.wikipedia.org/wiki/Precision_and_recall
        """
        return self.probability_of_detection()

    def precision(self) -> xr.DataArray:
        """
        Identical to the Success Ratio.

        What proportion of the forecast events actually eventuated?

        Returns:
            xr.DataArray: An xarray object containing the precision score

        .. math::
            \\text{precision} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false positives}}

        Notes:
            - Range: 0 to 1.  Perfect score: 1.
            - "True positives" is the same as "hits"
            - "False positives" is the same as "misses"

        References:
            - https://www.cawcr.gov.au/projects/verification/#SR
            - https://en.wikipedia.org/wiki/Precision_and_recall

        """
        return self.success_ratio()

    def f1_score(self) -> xr.DataArray:
        """
        Calculates the F1 score.

        Returns:
            xr.DataArray: An xarray object containing the F1 score

        .. math::
            \\text{F1} = \\frac{2 \\cdot \\text{true positives}}{(2 \\cdot  \\text{true positives}) +
            \\text{false positives} + \\text{false negatives}}

        Notes:
            - "True positives" is the same as "hits".
            - "False positives" is the same as "false alarms".
            - "False negatives" is the same as "misses".

        References:
            - https://en.wikipedia.org/wiki/F-score
        """
        cd = self.counts
        f1 = 2 * cd["tp_count"] / (2 * cd["tp_count"] + cd["fp_count"] + cd["fn_count"])
        return f1

    def equitable_threat_score(self) -> xr.DataArray:
        """
        Identical to the Gilbert Skill Score.

        Calculates the Equitable threat score.

        How well did the forecast "yes" events correspond to the observed "yes"
        events (accounting for hits due to chance)?

        Returns:
            xr.DataArray: An xarray object containing the equitable threat score

        .. math::
            \\text{ETS} = \\frac{\\text{true positives} - \\text{true positives} _\\text{random}}\
            {\\text{true positives} + \\text{false negatives} + \\text{false positives} - 
            \\text{true positives} _\\text{random}}

        where

        .. math::
            \\text{true_positives}_{\\text{random}} = \\frac{(\\text{true positives} + 
            \\text{false negatives}) (\\text{true positives} + \\text{false positives})}{\\text{total count}}

        Notes:
            - Range: -1/3 to 1, 0 indicates no skill. Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms"

        References:
            - Gilbert, G.K., 1884. Finley’s tornado predictions. American Meteorological Journal, 1(5), pp.166–172.
            - Hogan, R.J., Ferro, C.A., Jolliffe, I.T. and Stephenson, D.B., 2010. \
                Equitability revisited: Why the “equitable threat score” is not equitable. \
                Weather and Forecasting, 25(2), pp.710-726. https://doi.org/10.1175/2009WAF2222350.1
        """
        cd = self.counts
        hits_random = (cd["tp_count"] + cd["fn_count"]) * (cd["tp_count"] + cd["fp_count"]) / cd["total_count"]
        ets = (cd["tp_count"] - hits_random) / (cd["tp_count"] + cd["fn_count"] + cd["fp_count"] - hits_random)

        return ets

    def gilberts_skill_score(self) -> xr.DataArray:
        """
        Identical to the equitable threat scores

        Calculates the Gilbert skill score.

        How well did the forecast "yes" events correspond to the observed "yes"
        events (accounting for hits due to chance)?

        Returns:
            xr.DataArray: An xarray object containing the Gilberts Skill Score

        .. math::
            \\text{GSS} = \\frac{\\text{true positives} - \\text{true positives} _\\text{random}}\
            {\\text{true positives} + \\text{false negatives} + \\text{false positives} - 
            \\text{true positivies} _\\text{random}}

        where

        .. math::
            \\text{true_positives}_{\\text{random}} = \\frac{(\\text{true positives} + 
            \\text{false negatives}) (\\text{true positives} + \\text{false positives})}{\\text{total count}}

        Notes:
            - Range: -1/3 to 1, 0 indicates no skill. Perfect score: 1.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses"
            - "False positives" is the same as "false alarms".

        References:
            - Gilbert, G.K., 1884. Finley’s tornado predictions. American Meteorological Journal, 1(5), pp.166–172.
            - Hogan, R.J., Ferro, C.A., Jolliffe, I.T. and Stephenson, D.B., 2010. \
                Equitability revisited: Why the “equitable threat score” is not equitable. \
                Weather and Forecasting, 25(2), pp.710-726. https://doi.org/10.1175/2009WAF2222350.1
        """
        return self.equitable_threat_score()

    def heidke_skill_score(self) -> xr.DataArray:
        """
        Identical to Cohen's Kappa

        Calculates the Heidke skill score.

        What was the accuracy of the forecast relative to that of random chance?

        Returns:
            xr.DataArray: An xarray object containing the Heidke skill score

        .. math::
            \\\\
            \\text{HSS} =
                \\frac{\\text{true positives} + \\text{true negatives} - E_{random}}{
                       \\text{total count} - E_{\\text{random}}}

        where

        .. math::
            \\begin{aligned}
                E_{\\text{random}}
                    &= \\text{expected correct matches due to random chance}
                \\\\&= \\frac{\\left(\\text{tp} + \\text{fn}\\right) \\cdot
                              \\left(\\text{tp} + \\text{fp}\\right) +
                              \\left(\\text{tn} + \\text{fn}\\right) \\cdot
                              \\left(\\text{tn} + \\text{fp}\\right)}{
                              \\text{total count}}
            \\end{aligned}

        Notes:
            - Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
            - HSS = Heidke Skill Score
            - "True positives" (:math:`\\text{tp}`) is the same as "hits".
            - "False negatives" (:math:`\\text{fn}`) is the same as "misses".
            - "False positives" (:math:`\\text{fp}`) is the same as "false alarms".
            - "True negatives" (:math:`\\text{tn}`) is the same as "correct negatives".

        References:
            - https://en.wikipedia.org/wiki/Cohen%27s_kappa
        """
        cd = self.counts
        exp_correct = (1 / cd["total_count"]) * (
            (cd["tp_count"] + cd["fn_count"]) * (cd["tp_count"] + cd["fp_count"])
            + ((cd["tn_count"] + cd["fn_count"]) * (cd["tn_count"] + cd["fp_count"]))
        )
        hss = ((cd["tp_count"] + cd["tn_count"]) - exp_correct) / (cd["total_count"] - exp_correct)
        return hss

    def cohens_kappa(self) -> xr.DataArray:
        """
        Identical to the Heidke skill score.

        Calculates Cohen's kappa.

        What was the accuracy of the forecast relative to that of random chance?

        Returns:
            xr.DataArray: An xarray object containing the Cohen's Kappa score

        .. math::
            \\\\
            \\text{Cohen's Kappa} \\left(\\kappa\\right) =
                \\frac{\\text{true positives} + \\text{true negatives} - E_{random}}{
                       \\text{total count} - E_{\\text{random}}}

        where

        .. math::
            \\begin{aligned}
                E_{\\text{random}}
                    &= \\text{expected correct matches due to random chance}
                \\\\&= \\frac{\\left(\\text{tp} + \\text{fn}\\right) \\cdot
                              \\left(\\text{tp} + \\text{fp}\\right) +
                              \\left(\\text{tn} + \\text{fn}\\right) \\cdot
                              \\left(\\text{tn} + \\text{fp}\\right)}{
                              \\text{total count}}
            \\end{aligned}

        Notes:
            - Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
            - "True positives" (:math:`\\text{tp}`) is the same as "hits".
            - "False negatives" (:math:`\\text{fn}`) is the same as "misses".
            - "False positives" (:math:`\\text{fp}`) is the same as "false alarms".
            - "True negatives" (:math:`\\text{tn}`) is the same as "correct negatives".

        References:
            - https://en.wikipedia.org/wiki/Cohen%27s_kappa
        """
        return self.heidke_skill_score()

    def odds_ratio(self) -> xr.DataArray:
        """
        Calculates the odds ratio

        What is the ratio of the odds of a "yes" forecast being correct, to the odds of
        a "yes" forecast being wrong?

        Returns:
            xr.DataArray: An xarray object containing the odds ratio

        .. math::
            \\begin{aligned}
                \\text{odds ratio} &=
                    \\left[\\frac{\\text{POD}}{\\text{1 - POD}}\\right]
                        \\div
                    \\left[\\frac{\\text{POFD}}{\\text{1 - POFD}}\\right]
                              \\\\ &=
                    \\frac{\\text{true positives} \\cdot \\text{true negatives}}{\\text{false positives} \\cdot \\text{false negatives}}
            \\end{aligned}

        where

        .. math::
            \\text{POD} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false negatives}}

        and

        .. math::
            \\text{POFD} = \\frac{\\text{false positives}}{\\text{true negatives} + \\text{false positives}}


        Notes:
            - Range: 0 to ∞, 1 indicates no skill. Perfect score: ∞.
            - POD = Probability of Detection
            - POFD = Probability of False Detection
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            - Stephenson, D.B., 2000. Use of the “odds ratio” for diagnosing forecast skill. \
              Weather and Forecasting, 15(2), pp.221-232. \
              https://doi.org/10.1175/1520-0434(2000)015%3C0221:UOTORF%3E2.0.CO;2
        """
        odds_r = (self.probability_of_detection() / (1 - self.probability_of_detection())) / (
            self.probability_of_false_detection() / (1 - self.probability_of_false_detection())
        )
        return odds_r

    def odds_ratio_skill_score(self) -> xr.DataArray:
        """
        Identical to Yule's Q.

        Calculates the odds ratio skill score.

        What was the improvement of the forecast over random chance?

        Returns:
            xr.DataArray: An xarray object containing the odds ratio skill score

        .. math::

            \\begin{aligned}
                \\text{ORSS} &= \\frac{\\text{OR} - 1}{\\text{OR} + 1}
                        \\\\ &= \\frac{\\text{true positives} \\cdot \\text{true negatives}
                                         - \\text{false positives} \\cdot \\text{false negatives}}{
                                       \\text{true positives} \\cdot \\text{true negatives}
                                         + \\text{false positives} \\cdot \\text{false negatives}}
            \\end{aligned}

        Notes:
            - Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
            - ORSS = Odds Ratio Skill Score
            - OR = Odds ratio, see: :meth:`odds_ratio`
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            - Stephenson, D.B., 2000. Use of the “odds ratio” for diagnosing forecast skill. \
              Weather and Forecasting, 15(2), pp.221-232. \
              https://doi.org/10.1175/1520-0434(2000)015%3C0221:UOTORF%3E2.0.CO;2
        """
        cd = self.counts
        orss = (cd["tp_count"] * cd["tn_count"] - cd["fn_count"] * cd["fp_count"]) / (
            cd["tp_count"] * cd["tn_count"] + cd["fn_count"] * cd["fp_count"]
        )
        return orss

    def yules_q(self) -> xr.DataArray:
        """
        Identical to the odds ratio skill score.

        Calculates the Yule's Q.

        What was the improvement of the forecast over random chance?

        Returns:
            xr.DataArray: An xarray object containing Yule's Q

        .. math::

            \\begin{aligned}
                \\text{Yule's Q} &= \\frac{\\text{OR} - 1}{\\text{OR} + 1}
                        \\\\ &= \\frac{\\text{true positives} \\cdot \\text{true negatives}
                                         - \\text{false positives} \\cdot \\text{false negatives}}{
                                       \\text{true positives} \\cdot \\text{true negatives}
                                         + \\text{false positives} \\cdot \\text{false negatives}}
            \\end{aligned}

        Notes:
            - Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
            - OR = Odds ratio, see: :meth:`odds_ratio`.
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".

        References:
            Stephenson, D.B., 2000. Use of the “odds ratio” for diagnosing forecast skill. \
            Weather and Forecasting, 15(2), pp.221-232. \
            https://doi.org/10.1175/1520-0434(2000)015%3C0221:UOTORF%3E2.0.CO;2
        """
        return self.odds_ratio_skill_score()

    def symmetric_extremal_dependence_index(self) -> xr.DataArray:
        """
        Calculates the Symmetric Extremal Dependence Index (SEDI).

        Returns:
            xr.DataArray: An xarray object containing the SEDI score

        .. math::
            \\frac{\\ln(\\text{POFD}) - \\ln(\\text{POD}) + \\ln(\\text{1-POD}) - \\ln(\\text{1 -POFD})}
            {\\ln(\\text{POFD}) + \\ln(\\text{POD}) + \\ln(\\text{1-POD}) + \\ln(\\text{1 -POFD})}

        where

        .. math::
            \\text{POD} = \\frac{\\text{true positives}}{\\text{true positives} + \\text{false negatives}}

        and

        .. math::
            \\text{POFD} = \\frac{\\text{false positives}}{\\text{true negatives} + \\text{false positives}}


        Notes:
            - POD = Probability of Detection
            - POFD = Probability of False Detection
            - "True positives" is the same as "hits".
            - "False negatives" is the same as "misses".
            - "False positives" is the same as "false alarms".
            - "True negatives" is the same as "correct negatives".
            - Range: -1 to 1, Perfect score: 1.


        References:
            Ferro, C.A.T. and Stephenson, D.B., 2011. Extremal dependence indices: Improved verification
            measures for deterministic forecasts of rare binary events. Weather and Forecasting, 26(5), pp.699-713.
            https://doi.org/10.1175/WAF-D-10-05030.1
        """
        score = (
            np.log(self.probability_of_false_detection())
            - np.log(self.probability_of_detection())
            + np.log(1 - self.probability_of_detection())
            - np.log(1 - self.probability_of_false_detection())
        ) / (
            np.log(self.probability_of_false_detection())
            + np.log(self.probability_of_detection())
            + np.log(1 - self.probability_of_detection())
            + np.log(1 - self.probability_of_false_detection())
        )
        return score


class BinaryContingencyManager(BasicContingencyManager):
    """
    See https://scores.readthedocs.io/en/stable/tutorials/Binary_Contingency_Scores.html for
    a detailed walkthrough showing the use in practice.

    A BinaryContingencyManager holds the underlying binary forecast and observed event data,
    from which it builds a contingency table and provides scoring functionality based on that table.

    A BinaryContingencyManager is typically created by an :py:class:`EventOperator`, such as a
    :py:class:`ThresholdOperator`, but can also be created directly from binary data if the user
    wishes.

    Supported operations include:
        - "Transforming" the data in various ways, such as dimensional reduction
        - Producing contingency tables
        - Calculating scores and metrics based on contingency tables

    The full data comprises several n-dimensional binary arrays which can be considered maps of:
        - True positives (hits)
        - False positives (false alarms)
        - True negatives (correct negatives)
        - False negatives (misses)

    These masks, in addition to supporting score calculations, can be accessed and used for:

    - Plotting these attributes on a map
    - Masking these values by a geographical region prior to score calculation

    As such, the per-pixel information is useful as well as the overall ratios involved.

    BinaryContingencyManager inherits from (uses) the :py:class:`BasicContingencyManager` class to
    provide score calculations on the final contingency table. Documentation for the available scores
    is found in the :py:class:`BasicContingencyManager` API entry but the calls can be made directly
    against instances of BinaryContingencyManager where performance or transformation are not a concern.
    """

    def __init__(
        self, fcst_events: FlexibleArrayType, obs_events: FlexibleArrayType
    ):  # pylint: disable=super-init-not-called
        self.fcst_events = fcst_events
        self.obs_events = obs_events

        self.tp = (self.fcst_events == 1) & (self.obs_events == 1)  # true positives
        self.tn = (self.fcst_events == 0) & (self.obs_events == 0)  # true negatives
        self.fp = (self.fcst_events == 1) & (self.obs_events == 0)  # false positives
        self.fn = (self.fcst_events == 0) & (self.obs_events == 1)  # false negatives

        # Bring back NaNs where there is either a forecast or observed event nan
        self.tp = self.tp.where(~np.isnan(fcst_events))
        self.tp = self.tp.where(~np.isnan(obs_events))
        self.tn = self.tn.where(~np.isnan(fcst_events))
        self.tn = self.tn.where(~np.isnan(obs_events))
        self.fp = self.fp.where(~np.isnan(fcst_events))
        self.fp = self.fp.where(~np.isnan(obs_events))
        self.fn = self.fn.where(~np.isnan(fcst_events))
        self.fn = self.fn.where(~np.isnan(obs_events))

        # Variables for count-based metrics
        self.counts = self._get_counts()
        self._make_xr_table()

    def transform(
        self,
        *,
        reduce_dims: Optional[FlexibleDimensionTypes] = None,
        preserve_dims: Optional[FlexibleDimensionTypes] = None,
    ) -> BasicContingencyManager:
        """
        Compute the contingency table, preserving or reducing the specified dimensions.

        Args:
            - reduce_dims: Dimensions to reduce. Can be "all" to reduce all dimensions.
            - preserve_dims: Dimensions to preserve. Can be "all" to preserve all dimensions.

        Returns:
            scores.categorical.BasicContingencyManager: A `scores` class which supports efficient
            calculation of contingency metrics.
        """
        cd = self._get_counts(reduce_dims=reduce_dims, preserve_dims=preserve_dims)
        return BasicContingencyManager(cd)

    def _get_counts(
        self,
        *,
        reduce_dims: Optional[FlexibleDimensionTypes] = None,
        preserve_dims: Optional[FlexibleDimensionTypes] = None,
    ) -> dict:
        """
        Generates the uncomputed count values
        """

        to_reduce = scores.utils.gather_dimensions(
            self.fcst_events.dims,
            self.obs_events.dims,
            reduce_dims=reduce_dims,
            preserve_dims=preserve_dims,
        )

        cd = {
            "tp_count": self.tp.sum(dim=to_reduce),
            "tn_count": self.tn.sum(dim=to_reduce),
            "fp_count": self.fp.sum(dim=to_reduce),
            "fn_count": self.fn.sum(dim=to_reduce),
        }
        total = cd["tp_count"] + cd["tn_count"] + cd["fp_count"] + cd["fn_count"]
        cd["total_count"] = total

        return cd


class EventOperator(ABC):
    """
    Abstract Base Class (ABC) for event operators which can be used in deriving contingency
    tables. ABCs are not used directly but instead define the requirements for other classes
    and may be used for type checking.
    """

    @abstractmethod
    def make_event_tables(self, fcst: FlexibleArrayType, obs: FlexibleArrayType, *, event_threshold=None, op_fn=None):
        """
        This method should be over-ridden to return forecast and observed event tables
        """
        ...  # pragma: no cover # pylint: disable=unnecessary-ellipsis

    @abstractmethod
    def make_contingency_manager(
        self, fcst: FlexibleArrayType, obs: FlexibleArrayType, *, event_threshold=None, op_fn=None
    ):
        """
        This method should be over-ridden to return a contingency table.
        """
        ...  # pragma: no cover # pylint: disable=unnecessary-ellipsis


class ThresholdEventOperator(EventOperator):
    """
    See https://scores.readthedocs.io/en/stable/tutorials/Binary_Contingency_Scores.html for
    a detailed walkthrough showing the use in practice.

    A ThresholdEventOperator is used to produce a :py:class:`BinaryContingencyManager` from
    forecast and observed data. It considers an event to be defined when the forecast and
    observed variables meet a particular threshold condition (e.g. rainfall above 1mm).

    The class may be used for any variable, for any threshold, and for any comparison
    operator (e.g. greater-than, less-than, greater-than-or-equal-to, ... )
    """

    def __init__(self, *, precision=DEFAULT_PRECISION, default_event_threshold=0.001, default_op_fn=operator.ge):
        self.precision = precision
        self.default_event_threshold = default_event_threshold
        self.default_op_fn = default_op_fn

    def make_event_tables(self, fcst: FlexibleArrayType, obs: FlexibleArrayType, *, event_threshold=None, op_fn=None):
        """
        Using this function requires a careful understanding of the structure of the data
        and the use of the operator function. The default operator is a simple greater-than
        operator, so this will work on a simple DataArray. To work on a DataSet, a richer
        understanding is required. It is recommended to work through the Contingency Table
        tutorial to review more complex use cases, including on multivariate gridded model
        data, and on station data structures.
        """

        if not event_threshold:
            event_threshold = self.default_event_threshold

        if op_fn is None:
            op_fn = self.default_op_fn

        fcst_events = op_fn(fcst, event_threshold)
        obs_events = op_fn(obs, event_threshold)

        # Bring back NaNs
        fcst_events = fcst_events.where(~np.isnan(fcst))
        obs_events = obs_events.where(~np.isnan(obs))

        return (fcst_events, obs_events)

    def make_contingency_manager(
        self, fcst: FlexibleArrayType, obs: FlexibleArrayType, *, event_threshold=None, op_fn=None
    ) -> BinaryContingencyManager:
        """
        Using this function requires a careful understanding of the structure of the data
        and the use of the operator function. The default operator is a simple greater-than
        operator, so this will work on a simple DataArray. To work on a DataSet, a richer
        understanding is required. It is recommended to work through the Contingency Table
        tutorial to review more complex use cases, including on multivariate gridded model
        data, and on station data structures.
        """

        if not event_threshold:
            event_threshold = self.default_event_threshold

        if op_fn is None:
            op_fn = self.default_op_fn

        fcst_events = op_fn(fcst, event_threshold)
        obs_events = op_fn(obs, event_threshold)

        # Bring back NaNs
        fcst_events = fcst_events.where(~np.isnan(fcst))
        obs_events = obs_events.where(~np.isnan(obs))

        table = BinaryContingencyManager(fcst_events, obs_events)
        return table



"""
This module contains functions to calculate the correlation.
"""

from typing import Optional

import xarray as xr

import scores.utils
from scores.typing import FlexibleDimensionTypes


def pearsonr(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
) -> xr.DataArray:
    """
    Calculates the Pearson's correlation coefficient between two xarray DataArrays

    .. math::
        \\rho = \\frac{\\sum_{i=1}^{n}{(x_i - \\bar{x})(y_i - \\bar{y})}}
        {\\sqrt{\\sum_{i=1}^{n}{(x_i-\\bar{x})^2}\\sum_{i=1}^{n}{(y_i - \\bar{y})^2}}}

    where:
        - :math:`\\rho` = Pearson's correlation coefficient
        - :math:`x_i` = the values of x in a sample (i.e. forecast values)
        - :math:`\\bar{x}` = the mean value of the forecast sample
        - :math:`y_i` = the values of y in a sample (i.e. observed values)
        - :math:`\\bar{y}` = the mean value of the observed sample value

    Args:
        fcst: Forecast or predicted variables
        obs: Observed variables.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the Pearson's correlation coefficient.
            All other dimensions will be preserved.
        preserve_dims: Optionally specify which dimensions to preserve when
            calculating the Pearson's correlation coefficient. All other dimensions will
            be reduced. As a special case, 'all' will allow all dimensions to be
            preserved. In this case, the result will be in the same shape/dimensionality
            as the forecast, and the errors will be the absolute error at each
            point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely.
    Returns:
        xr.DataArray: An xarray object with Pearson's correlation coefficient values

    Note:
        This function isn't set up to take weights.

    Reference:
        https://en.wikipedia.org/wiki/Pearson_correlation_coefficient
    """
    reduce_dims = scores.utils.gather_dimensions(
        fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims
    )

    return xr.corr(fcst, obs, reduce_dims)



"""
Implementation of quantile loss (score)
"""
from typing import Optional

import xarray as xr

from scores.functions import apply_weights
from scores.typing import FlexibleDimensionTypes, XarrayLike
from scores.utils import check_dims, gather_dimensions


def quantile_score(
    fcst: XarrayLike,
    obs: XarrayLike,
    alpha: float,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[XarrayLike] = None,
) -> XarrayLike:
    """
    Calculates a score that targets alpha-quantiles.
    Use with alpha = 0.5 for forecasts of the median.
    Use with alpha = 0.9 for forecasts of the 90th percentile.

    Args:
        fcst: array of forecasts
        obs: array of observations
        alpha: A value between 0 and 1 (exclusive)
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the quantile score. All other dimensions will be preserved. As a
            special case, 'all' will allow all dimensions to be reduced. Only one
            of `reduce_dims` and `preserve_dims` can be supplied. The default behaviour
            if neither are supplied is to reduce all dims.
        preserve_dims: Optionally specify which dimensions to preserve when calculating
            quantile score. All other dimensions will be reduced. As a special case, 'all'
            will allow all dimensions to be preserved. In this case, the result will be in
            the same shape/dimensionality as the forecast, and the errors will be the quantile
            score at each point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely. Only one of `reduce_dims`
            and `preserve_dims` can be supplied. The default behaviour if neither are supplied
            is to reduce all dims.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)

    Returns:
        A DataArray with values being the mean generalised piecewise linear (GPL)
        scoring function, with the dimensions specified in `dims`.
        If `dims` is `None`, the returned DataArray will have only one element,
        the overall mean GPL score.

    Raises:
        ValueError: if `alpha` is not between 0 and 1.

    Notes:

        .. math::

            gpl(x) = \\begin{cases}\\alpha * (-x) & x \\leq 0\\\\
           (1-\\alpha) x & x > 0\\end{cases}

        where:

            - :math:`\\alpha` is the targeted quantile.
            - :math:`x` is the difference, fcst - obs

    References:
        T. Gneiting, "Making and evaluating point forecasts",
        J. Amer. Stat. Assoc., Vol. 106 No. 494 (June 2011), pp. 754--755,
        Theorem 9

    """
    specified_dims = reduce_dims or preserve_dims
    # check requested dims are a subset of fcst dimensions
    if specified_dims is not None:
        check_dims(xr_data=fcst, expected_dims=specified_dims, mode="superset")
    # check obs dimensions are a subset of fcst dimensions
    check_dims(xr_data=obs, expected_dims=fcst.dims, mode="subset")  # type: ignore

    # check that alpha is between 0 and 1 as required
    if (alpha <= 0) or (alpha >= 1):
        raise ValueError("alpha is not between 0 and 1")

    # Do this operation once to save compute time
    diff = fcst - obs

    # calculate the score applicable when fcst <= obs
    score_fcst_lte_obs = alpha * (-diff)

    # calculate the score applicable when fcst > obs
    score_fcst_ge_obs = (1 - alpha) * diff

    result = xr.where(diff > 0, score_fcst_ge_obs, score_fcst_lte_obs)

    reduce_dims = gather_dimensions(fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims)  # type: ignore[assignment]
    results = apply_weights(result, weights=weights)
    score = results.mean(dim=reduce_dims)

    return score  # type: ignore



"""
Convenience module to retrieve compute backend
"""
from scores.fast.fss.fss_numba import FssNumba
from scores.fast.fss.fss_numpy import FssNumpy
from scores.fast.fss.typing import FssComputeMethod


def get_compute_backend(compute_method: FssComputeMethod):  # pragma: no cover
    """
    Returns the appropriate compute backend class constructor.
    """
    # Note: compute methods other than NUMPY currently not implemented
    if compute_method == FssComputeMethod.NUMPY:
        return FssNumpy
    if compute_method == FssComputeMethod.NUMBA:
        # Note: currently not implemented fore demonstration purposes only
        # this should throw a compatiblity error.
        return FssNumba
    raise NotImplementedError(f"Compute method {compute_method} not implemented.")



"""
Backend for computing Fractions Skill Score (FSS) using numpy
"""

from dataclasses import dataclass

import numpy as np

from scores.fast.fss.backend import FssBackend
from scores.fast.fss.typing import FssComputeMethod, FssDecomposed, f8x3

_COMPATIBLE = True


@dataclass
class FssNumpy(FssBackend):
    """
    Implementation of numpy backend for computing FSS
    """

    compute_method = FssComputeMethod.NUMPY

    def _check_compatibility(self):  # pragma: no cover
        """
        Numpy should be compatible by default.
        """
        # false branch - should always return
        # kept here to preserve the pattern consistency with other backends
        if _COMPATIBLE:
            return
        # unreachable
        super()._check_compatibility()

    def _compute_integral_field(self):  # pylint: disable=too-many-locals
        obs_partial_sums = self._obs_pop.cumsum(1).cumsum(0)
        fcst_partial_sums = self._fcst_pop.cumsum(1).cumsum(0)

        # pad first row/column with 0s, addressses off-by-one window sizes for the first row/column
        # due to the way meshgrid indexing works.
        zero_row = np.zeros((fcst_partial_sums.shape[0], 1))
        obs_partial_sums = np.hstack([zero_row, obs_partial_sums])
        fcst_partial_sums = np.hstack([zero_row, fcst_partial_sums])

        zero_col = np.zeros((1, fcst_partial_sums.shape[1]))
        obs_partial_sums = np.vstack([zero_col, obs_partial_sums])
        fcst_partial_sums = np.vstack([zero_col, fcst_partial_sums])

        if self.zero_padding:
            # ------------------------------
            # 0-padding
            # ------------------------------
            #    ..................
            #    ..................
            #    ..+------------+..
            #    ..|            |..
            #    ..|            |..
            #    ..|            |..
            #    ..+------------+..
            #    ..................
            #    ..................
            #
            # ".." represents 0 padding
            # the rectangle represents the
            # FSS computation region.
            # ------------------------------
            (w_h, w_w) = (self.window_size[0], self.window_size[1])
            half_w_h = int(w_h / 2)
            half_w_w = int(w_w / 2)
            im_h = fcst_partial_sums.shape[0]
            im_w = fcst_partial_sums.shape[1]

            # get remaining window size, incase window size is odd
            rem_h = w_h - half_w_h
            rem_w = w_w - half_w_w

            # Zero padding is equivilent to clamping the coordinate values at the border
            # r = rows, c = cols, tl = top-left, br = bottom-right
            r_tl = np.clip(np.arange(-half_w_h, im_h - half_w_h), 0, im_h - 1)
            c_tl = np.clip(np.arange(-half_w_w, im_w - half_w_w), 0, im_w - 1)
            mesh_tl = np.meshgrid(r_tl, c_tl, indexing="ij")

            r_br = np.clip(np.arange(rem_h, im_h + rem_h), 1, im_h - 1)
            c_br = np.clip(np.arange(rem_w, im_w + rem_w), 1, im_w - 1)
            mesh_br = np.meshgrid(r_br, c_br, indexing="ij")

        else:
            # ------------------------------
            # interior border
            # ------------------------------
            # data at the border is trim
            # -ed and used for padding
            #    +---------------+
            #    | +-----------+ |
            #    | |///////////| |
            #    | |///////////| |
            #    | |///////////| |
            #    | +-----------+ |
            #    +---------------+
            #
            # "//" represents the effective
            # FSS computation region
            # ------------------------------
            im_h = fcst_partial_sums.shape[0] - self.window_size[0]
            im_w = fcst_partial_sums.shape[1] - self.window_size[1]
            mesh_tl = np.mgrid[0:im_h, 0:im_w]
            mesh_br = (mesh_tl[0] + self.window_size[0], mesh_tl[1] + self.window_size[1])

        # ----------------------------------
        # Computing window area from sum
        # area table
        # ----------------------------------
        #     A          B
        # tl.0,tl.1    tl.0,br.1
        #     +----------+
        #     |          |
        #     |          |
        #     |          |
        #     +----------+
        # br.0,tl.1   br.0,br.1
        #     C          D
        #
        # area = D - B - C + A
        # -----------------------------------

        obs_a = obs_partial_sums[mesh_tl[0], mesh_tl[1]]
        obs_b = obs_partial_sums[mesh_tl[0], mesh_br[1]]
        obs_c = obs_partial_sums[mesh_br[0], mesh_tl[1]]
        obs_d = obs_partial_sums[mesh_br[0], mesh_br[1]]
        self._obs_img = obs_d - obs_b - obs_c + obs_a

        fcst_a = fcst_partial_sums[mesh_tl[0], mesh_tl[1]]
        fcst_b = fcst_partial_sums[mesh_tl[0], mesh_br[1]]
        fcst_c = fcst_partial_sums[mesh_br[0], mesh_tl[1]]
        fcst_d = fcst_partial_sums[mesh_br[0], mesh_br[1]]
        self._fcst_img = fcst_d - fcst_b - fcst_c + fcst_a

        return self

    def _compute_fss_components(self) -> FssDecomposed:
        diff = np.nanmean(np.power(self._obs_img - self._fcst_img, 2))
        fcst = np.nanmean(np.power(self._fcst_img, 2))
        obs = np.nanmean(np.power(self._obs_img, 2))
        return np.void((fcst, obs, diff), dtype=f8x3)



"""
Murphy score
"""
from collections.abc import Sequence
from typing import Any, Callable, Literal, Optional, Union

import numpy as np
import xarray as xr

from scores.processing import broadcast_and_match_nan
from scores.typing import FlexibleArrayType, FlexibleDimensionTypes
from scores.utils import gather_dimensions

QUANTILE = "quantile"
HUBER = "huber"
EXPECTILE = "expectile"
VALID_SCORING_FUNC_NAMES = [QUANTILE, HUBER, EXPECTILE]
SCORING_FUNC_DOCSTRING_PARAMS = {fun.upper(): fun for fun in VALID_SCORING_FUNC_NAMES}


def murphy_score(  # pylint: disable=R0914
    fcst: xr.DataArray,
    obs: xr.DataArray,
    thetas: Union[Sequence[float], xr.DataArray],
    *,  # Force keywords arguments to be keyword-only
    functional: Literal["quantile", "huber", "expectile"],
    alpha: float,
    huber_a: Optional[float] = None,
    decomposition: bool = False,
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
) -> xr.Dataset:
    """Returns the mean elementary score (Ehm et. al. 2016), also known as Murphy score,
    evaluated at decision thresholds specified by thetas. Optionally returns a decomposition
    of the score in terms of penalties for over- and under-forecasting.

    Select `functional="quantile"` and `alpha=0.5` for the median functional.
    Select `functional="expectile"` and `alpha=0.5` for the mean (i.e., expectation) functional.

    Consider using `murphy_thetas` to generate thetas. If utilising dask, you may want
    to store `thetas` as a netCDF on disk and pass it in as an xarray object. Otherwise,
    very large objects may be created when `fcst`, `obs` and `thetas` are broadcast
    together.


    Args:
        fcst: Forecast numerical values.
        obs: Observed numerical values.
        thetas: Theta thresholds.
        functional: The type of elementary scoring function, one of {QUANTILE},
            {HUBER}, or {EXPECTILE}.
        alpha: Risk parameter (i.e. quantile or expectile level) for the functional. Must be between 0 and 1.
        huber_a: Huber transition parameter, used for {HUBER} functional only.
            Must be strictly greater than 0.
        decomposition: True to return penalty values due to under- and over-fcst
            as well as the total score, False to return total score only.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the Murphy score. All other dimensions will be preserved. As a
            special case, 'all' will allow all dimensions to be reduced. Only one
            of `reduce_dims` and `preserve_dims` can be supplied. The default behaviour
            if neither are supplied is to reduce all dims.
        preserve_dims: Optionally specify which dimensions to preserve
            when calculating the Murphy score. All other dimensions will be reduced.
            As a special case, 'all' will allow all dimensions to be
            preserved. In this case, the result will be in the same
            shape/dimensionality as the forecast, and the errors will be
            the FIRM score at each point (i.e. single-value comparison
            against observed), and the forecast and observed dimensions
            must match precisely. Only one of `reduce_dims` and `preserve_dims` can be
            supplied. The default behaviour if neither are supplied is to reduce all dims.

    Returns:
        An xr.Dataset with dimensions based on the `preserve_dims` or `reduce_dims` arg
            as  well as a "theta" dimension with values matching `thetas` input.

        If `decomposition` is False, the dataset's variables will contain 1 element
            "total".

        If `decomposition` is True, in addition to "total", it will have
            "underforecast" and "overforecast" data_vars.

    Raises:
        ValueError: If `functional` is not one of the expected functions.
        ValueError: If `alpha` is not strictly between 0 and 1.
        ValueError: If `huber_a` is not strictly greater than 0.

    References:
        For mean elementary score definitions, see

            - Theorem 1 of Ehm et. al. (2016) "Of Quantiles and Expectiles", J. Royal Stat. Soc. B,
              78(3): 505–562. https://www.jstor.org/stable/24775351.
            - Theorem 5.3 of Taggart (2022) "Point forecasting and forecast evaluation with
              generalized Huber loss", Electron. J. Statist. 16(1): 201-231.
              DOI: 10.1214/21-EJS1957

    """
    functional_lower = functional.lower()
    _check_murphy_inputs(alpha=alpha, functional=functional_lower, huber_a=huber_a)
    if isinstance(thetas, xr.DataArray):
        theta1 = thetas
    else:
        theta1 = xr.DataArray(data=thetas, dims=["theta"], coords={"theta": thetas})
    theta1, fcst1, obs1 = broadcast_and_match_nan(theta1, fcst, obs)  # type: ignore

    over, under = exposed_functions()[f"_{functional_lower}_elementary_score"](
        fcst1, obs1, theta1, alpha, huber_a=huber_a
    )
    # Align dimensions, this is required in cases such as when the station numbers
    # are not in the same order in `obs` and `fcst` to prevent an exception on the next
    # line that combines the scores
    over, under, fcst1, *_ = xr.align(over, under, fcst1)
    score = over.combine_first(under).fillna(0).where(~np.isnan(fcst1), np.nan)

    sources = [score]
    names = ["total"]
    if decomposition:
        over = over.fillna(0).where(~np.isnan(fcst1), np.nan)
        under = under.fillna(0).where(~np.isnan(fcst1), np.nan)
        sources += [under, over]
        names += ["underforecast", "overforecast"]
    for source, name in zip(sources, names):
        source.name = name
    result = xr.merge(sources)
    reduce_dims = gather_dimensions(fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims)
    result = result.mean(dim=reduce_dims)
    return result


def _quantile_elementary_score(fcst: FlexibleArrayType, obs: FlexibleArrayType, theta, alpha, **_):
    """Return over and under forecast vs obs penalties relative to theta for {QUANTILE}."""
    zero_array = fcst * 0.0
    over = (zero_array + 1 - alpha).where((fcst > theta) & (obs <= theta))
    under = (zero_array + alpha).where((fcst <= theta) & (obs > theta))
    return over, under


def _huber_elementary_score(fcst: FlexibleArrayType, obs: FlexibleArrayType, theta, alpha, *, huber_a):
    """Return over and under forecast vs obs penalties relative to theta for {HUBER}."""
    zero_array = fcst * 0.0
    over = (1 - alpha) * np.minimum(theta - obs, zero_array + huber_a).where((fcst > theta) & (obs <= theta))
    under = alpha * np.minimum(obs - theta, zero_array + huber_a).where((fcst <= theta) & (obs > theta))
    return over, under


def _expectile_elementary_score(fcst: FlexibleArrayType, obs: FlexibleArrayType, theta, alpha, **_):
    """Return over and under forecast vs obs penalties relative to theta for {EXPECTILE}."""
    over = (1 - alpha) * np.abs(obs - theta).where((fcst > theta) & (obs <= theta))
    under = alpha * np.abs(obs - theta).where((fcst <= theta) & (obs > theta))
    return over, under


def _check_murphy_inputs(*, alpha=None, functional=None, huber_a=None, left_limit_delta=None):
    """Raise ValueError if the arguments have unexpected values."""
    if (alpha is not None) and not (0 < alpha < 1):  # pylint: disable=C0325
        err = f"alpha (={alpha}) argument for Murphy scoring function should be strictly " "between 0 and 1."
        raise ValueError(err)
    if (functional is not None) and (functional not in VALID_SCORING_FUNC_NAMES):
        err = (
            f"Functional option '{functional}' for Murphy scoring function is "
            f"unknown, should be one of {VALID_SCORING_FUNC_NAMES}."
        )
        raise ValueError(err)
    if (functional == HUBER) and ((huber_a is None) or (huber_a <= 0)):
        err = f"huber_a (={huber_a}) argument should be > 0 when functional='{HUBER}'."
        raise ValueError(err)
    if (left_limit_delta is not None) and (left_limit_delta < 0):
        err = f"left_limit_delta (={left_limit_delta}) argument should be >= 0."
        raise ValueError(err)


def murphy_thetas(
    forecasts: list[xr.DataArray],
    obs: xr.DataArray,
    functional: Literal["quantile", "huber", "expectile"],
    *,  # Force keywords arguments to be keyword-only
    huber_a: Optional[float] = None,
    left_limit_delta: Optional[float] = None,
) -> list[float]:
    """Return the decision thresholds (theta values) at which to evaluate
    elementary scores for the construction of Murphy diagrams.

    This function may generate a very large number of theta thresholds if the forecast
    and obs data arrays are large and their values have high precision, which may lead
    to long computational times for Murphy diagrams. To reduce the number of thetas,
    users can first round forecast and obs values to an appropriate resolution.

    Args:
        forecasts: Forecast values, one array per source.
        obs: Observed values.
        functional: The type of elementary scoring function, one of {QUANTILE},
            {HUBER}, or {EXPECTILE}.
        huber_a: Huber transition parameter, used for {HUBER} functional only.
            Must be strictly greater than 0.
        left_limit_delta: Approximation of the left hand limit, used for {HUBER}
            and {EXPECTILE} functionals. Must be greater than or equal to 0.
            None will be treated as 0. Ideally, left_limit_delta should be
            small relative to the fcst and obs precision, and not greater than
            that precision.

    Returns:
        List[float]: theta thresholds to be used to compute murphy scores.

    Raises:
        ValueError: If `functional` is not one of the expected functions.
        ValueError: If `huber_a` is not strictly greater than 0.
        ValueError: If `left_limit_delta` is not greater than or equal to 0.

    Notes:
        For theta values at which evaluate elementary scores, see

        - Corollary 2 (p.521) of Ehm et. al. (2016) "Of Quantiles and Expectiles", J. Royal Stat. Soc. B,
          78(3): 505–562. https://www.jstor.org/stable/24775351.
        - Corollary 5.6 of Taggart (2022) "Point forecasting and forecast evaluation with
          generalized Huber loss", Electron. J. Statist. 16(1): 201-231.
          DOI: 10.1214/21-EJS1957

    """
    _check_murphy_inputs(functional=functional, huber_a=huber_a, left_limit_delta=left_limit_delta)
    if (left_limit_delta is None) and (functional in ["huber", "expectile"]):
        left_limit_delta = 0

    func = exposed_functions()[f"_{functional}_thetas"]
    result = func(
        forecasts=forecasts,
        obs=obs,
        huber_a=huber_a,
        left_limit_delta=left_limit_delta,
    )
    return result  # type: ignore


def _quantile_thetas(forecasts, obs, **_):
    """Return thetas for {QUANTILE} elementary scoring function."""
    ufcasts_and_uobs = np.unique(np.concatenate([*[x.values.flatten() for x in forecasts], obs.values.flatten()]))
    result = ufcasts_and_uobs[~np.isnan(ufcasts_and_uobs)]
    return list(result)


def _huber_thetas(forecasts, obs, *, huber_a, left_limit_delta, **_):
    """Return thetas for {HUBER} elementary scoring function."""
    uobs = np.unique(obs)
    uobs_minus_a = uobs - huber_a
    uobs_plus_a = uobs + huber_a
    ufcasts = np.unique(forecasts)
    left_limit_points = ufcasts - left_limit_delta
    ufcasts_and_uobs = np.unique(np.concatenate([ufcasts, left_limit_points, uobs, uobs_minus_a, uobs_plus_a]))
    result = ufcasts_and_uobs[~np.isnan(ufcasts_and_uobs)]
    return list(result)


def _expectile_thetas(forecasts, obs, *, left_limit_delta, **_):
    """Return thetas for {EXPECTILE} elementary scoring function."""
    ufcasts = np.unique(forecasts)
    left_limit_points = ufcasts - left_limit_delta
    ufcasts_and_uobs = np.unique(np.concatenate([ufcasts, left_limit_points, obs.values.flatten()]))
    result = ufcasts_and_uobs[~np.isnan(ufcasts_and_uobs)]
    return list(result)


def exposed_functions() -> dict[str, Callable[..., Any]]:
    """Expose functions used in calculation for easy usage upon user arguments."""
    return {
        "_quantile_elementary_score": _quantile_elementary_score,
        "_huber_elementary_score": _huber_elementary_score,
        "_expectile_elementary_score": _expectile_elementary_score,
        "_quantile_thetas": _quantile_thetas,
        "_huber_thetas": _huber_thetas,
        "_expectile_thetas": _expectile_thetas,
    }



"""
This module is currently empty.

This module is intended to hold code for new metrics that may not yet have a publication
and may not yet be in their final form, but which still have interest and value to
the verification community.

To propose a metric for inclusion in this API, please either get in touch through the
"Discussions" area of the github page (https://github.com/nci/scores/discussions) or
raise a new issue (https://github.com/nci/scores/issues)
"""



"""
Import the functions from the implementations into the public API
"""

from scores.spatial.fss_impl import fss_2d, fss_2d_binary, fss_2d_single_field

__all__ = ["fss_2d", "fss_2d_binary", "fss_2d_single_field"]



"""
Backend for computing Fractions Skill Score (FSS) using numba
    
"""
from dataclasses import dataclass

from scores.fast.fss.backend import FssBackend
from scores.fast.fss.typing import FssComputeMethod

_COMPATIBLE = True
try:
    import numba  # pylint: disable=unused-import
except ImportError:  # pragma: no cover
    _COMPATIBLE = False


@dataclass
class FssNumba(FssBackend):  # pragma: no cover
    """
    Implementation of numba backend for computing FSS

    This is currently a stub.
    """

    compute_method = FssComputeMethod.NUMBA

    def _check_compatibility(self):
        """
        Currently not implemented.
        """
        if _COMPATIBLE:
            return
        super()._check_compatibility()

    def _compute_integral_field(self):
        super()._compute_integral_field()

    def _compute_fss_components(self):
        super()._compute_integral_field()



"""
Backend implementation of the FSS algorithm. Currently calculations are
performed mainly by the `numpy` backend.
"""

from dataclasses import dataclass, field
from typing import Tuple

import numpy as np
import numpy.typing as npt

from scores.fast.fss.typing import FssComputeMethod, FssDecomposed
from scores.utils import BinaryOperator, CompatibilityError, DimensionError


@dataclass
class FssBackend:  # pylint: disable=too-many-instance-attributes
    """
    Abstract base class for computing fss.

    required methods:
        - :py:meth:`_compute_fss_components`
        - :py:meth:`_compute_integral_field`
    """

    compute_method = FssComputeMethod.INVALID

    fcst: npt.NDArray[np.float64]
    obs: npt.NDArray[np.float64]
    # KW_ONLY in dataclasses only supported for python >=3.10
    # _: KW_ONLY
    event_threshold: np.float64
    threshold_operator: BinaryOperator
    window_size: Tuple[int, int]
    zero_padding: bool

    # internal buffers
    _obs_pop: npt.NDArray[np.int64] = field(init=False)
    _fcst_pop: npt.NDArray[np.int64] = field(init=False)
    _obs_img: npt.NDArray[np.int64] = field(init=False)
    _fcst_img: npt.NDArray[np.int64] = field(init=False)

    def __post_init__(self):
        """
        Post-initialization checks go here.
        """
        self._check_dims()
        self._check_compatibility()

    def _check_compatibility(self):  # pragma: no cover
        raise CompatibilityError(
            f"Unable to run `{self.compute_method}`. Are you missing extra/optional dependencies in your install?"
        )

    def _check_dims(self):
        if self.fcst.shape != self.obs.shape:
            raise DimensionError("fcst and obs shapes do not match")
        if (
            self.window_size[0] > self.fcst.shape[0]
            or self.window_size[1] > self.fcst.shape[1]
            or self.window_size[0] < 1
            or self.window_size[1] < 1
        ):
            raise DimensionError(
                "invalid window size, `window_size` must be smaller than input data shape and greater than 0"
            )

    def _apply_event_threshold(self):
        """
        Default implementation of converting the input fields into binary fields by comparing
        against a event threshold; using numpy. This is a relatively cheap operation, so a specific
        implementation in derived backends is optional.
        """
        _op = self.threshold_operator.get()
        self._obs_pop = _op(self.obs, self.event_threshold)
        self._fcst_pop = _op(self.fcst, self.event_threshold)
        return self

    def _compute_integral_field(self):
        """
        Computes the rolling windowed sums over the entire observation & forecast fields. The
        resulting "integral field (or image)" can be cached in `self._obs_img` for observations, and
        `self._fcst_img` for forecast.

        See also: https://en.wikipedia.org/wiki/Summed-area_table

        Note: for non-`python` backends e.g. `rust`, its often the case that the integral field is
        computed & cached in the native language. In which case this method can just be overriden
        with a `return self`.
        """
        raise NotImplementedError(f"_compute_integral_field not implemented for {self.compute_method}")

    def _compute_fss_components(self) -> FssDecomposed:
        """
        FSS is roughly defined as
        ```
            fss = 1 - sum_w((p_o - p_f)^2) / (sum_w(p_o^2) + sum_w(p_f^2))

            where,
            p_o: observation populace > event_threshold, in one window
            p_f: forecast populace > event_threshold, in one window
            sum_w: sum over all windows
        ````

        In order to accumulate scores over non spatial dimensions at a higher level operation, we
        need to keep track of the de-composed sums as they need to be accumulated separately.

        The "fss components", hence, consist of:
        - `power_diff :: float = sum_w((p_o - p_f)^2)`
        - `power_obs :: float = sum_w(p_o^2)`
        - `power_fcst :: float = sum_w(p_f^2)`

        **WARNING:** returning sums of powers can result in overflows on aggregation. It is advisable
        to return the means instead, as it is equivilent with minimal computational trade-off.

        Returns:
            Tuple[float, float, float]: (power_fcst, power_obs, power_diff) as described above
        """
        raise NotImplementedError(f"`_compute_fss_components` not implemented for {self.compute_method}")

    def compute_fss_decomposed(self) -> FssDecomposed:
        """
        Calls the main pipeline to compute the fss score.

        Returns the decomposed scores for aggregation. This is because aggregation cannot be
        done on the final score directly.

        Note: FSS computations of stand-alone 2-D fields should use :py:meth:`compute_fss` instead.
        """
        # fmt: off
        return (
            self._apply_event_threshold()  # pylint: disable=protected-access
                ._compute_integral_field()
                ._compute_fss_components()
        )
        # fmt: on

    def compute_fss(self) -> np.float64:
        """
        Uses the components from :py:method:`compute_fss_decomposed` to compute the final
        Fractions Skill Score.
        """
        (fcst, obs, diff) = self.compute_fss_decomposed()
        denom = fcst + obs
        fss: np.float = 0.0

        if denom > 0.0:
            fss = 1.0 - diff / denom

        fss_clamped = max(min(fss, 1.0), 0.0)

        return fss_clamped



"""
Specific type definitions for Fss backend    
"""
from enum import Enum
from typing import TYPE_CHECKING, Union

import numpy as np
import numpy.typing as npt

# Note: soft keyword `type` only support from >=3.10
# "tuple" of 64 bit floats
f8x3 = np.dtype("f8, f8, f8")
# Note: `TypeAlias` on variables not available for python <3.10
if TYPE_CHECKING:  # pragma: no cover
    FssDecomposed = Union[np.ArrayLike, np.DtypeLike]
else:  # pragma: no cover
    FssDecomposed = npt.NDArray[f8x3]


class FssComputeMethod(Enum):
    """
    Choice of compute backend for FSS.
    """

    #: invalid backend
    INVALID = -1
    #: computations using numpy (default)
    NUMPY = 1
    #: computations using numba (currently unimplemented)
    NUMBA = 2



"""
Import the functions from the implementations into the public API
"""

from scores.processing.cdf.cdf_functions import (
    add_thresholds,
    cdf_envelope,
    decreasing_cdfs,
    fill_cdf,
    integrate_square_piecewise_linear,
    observed_cdf,
    propagate_nan,
    round_values,
)

__all__ = [
    "round_values",
    "propagate_nan",
    "observed_cdf",
    "integrate_square_piecewise_linear",
    "add_thresholds",
    "fill_cdf",
    "decreasing_cdfs",
    "cdf_envelope",
]



"""
Implementation of scoring functions that are consistent for
single-valued forecasts targeting quantiles, expectiles or Huber functionals.
"""

from typing import Callable, Optional

import xarray as xr

from scores.functions import apply_weights
from scores.typing import FlexibleDimensionTypes
from scores.utils import gather_dimensions


def consistent_expectile_score(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    alpha: float,
    phi: Callable[[xr.DataArray], xr.DataArray],
    phi_prime: Callable[[xr.DataArray], xr.DataArray],
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[xr.DataArray] = None,
) -> xr.DataArray:
    """
    Calculates the score using a scoring function that is consistent for the
    alpha-expectile functional, based on a supplied convex function phi.
    See Geniting (2011), or Equation (10) from Taggart (2022).

    .. math::

        S(x, y) =
        \\begin{cases}
        (1 - \\alpha)(\\phi(y) - \\phi(x) - \\phi'(x)(y-x)), & y < x \\\\
        \\alpha(\\phi(y) - \\phi(x) - \\phi'(x)(y-x)), & x \\leq y
        \\end{cases}

    where
        - :math:`x` is the forecast
        - :math:`y` is the observation
        - :math:`\\alpha` is the expectile level
        - :math:`\\phi` is a convex function of a single variable
        - :math:`\\phi'` is the subderivative of :math:`\\phi`
        - :math:`S(x,y)` is the score.

    Note that if :math:`\\phi` is differentiable then :math:`\\phi'` is its derivative.

    The score is negatively oriented with :math:`0 \\leq S(x,y) < \\infty`.

    We strongly recommend that you work through the tutorial 
    https://scores.readthedocs.io/en/stable/tutorials/Consistent_Scores.html to help you 
    understand how to use the consistent scoring functions in ``scores``.

    Args:
        fcst: array of forecast values.
        obs: array of corresponding observation values.
        alpha: expectile level. Must be strictly between 0 and 1.
        phi: a convex function on the real numbers, accepting a single array like argument.
        phi_prime: a subderivative of `phi`, accepting a single array like argument.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the consistent expectile score. All other dimensions will be preserved. As a
            special case, 'all' will allow all dimensions to be reduced. Only one
            of `reduce_dims` and `preserve_dims` can be supplied. The default behaviour
            if neither are supplied is to reduce all dims.
        preserve_dims: Optionally specify which dimensions to preserve when calculating
            the consistent quantile score. All other dimensions will be reduced. As a special case, 'all'
            will allow all dimensions to be preserved. In this case, the result will be in
            the same shape/dimensionality as the forecast, and the errors will be the consistent quantile
            score at each point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely. Only one of `reduce_dims`
            and `preserve_dims` can be supplied. The default behaviour if neither are supplied
            is to reduce all dims.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)

    Returns:
        array of (mean) scores that is consistent for alpha-expectile functional,
        with the dimensions specified by `dims`. If `dims` is `None`, the returned DataArray will have
        only one entry, the overall mean score.

    Raises:
        ValueError: if `alpha` is not strictly between 0 and 1.

    References:
        -   Gneiting, T. (2011). Making and Evaluating Point Forecasts. Journal of the 
            American Statistical Association, 106(494), 746–762. 
            https://doi.org/10.1198/jasa.2011.r10138
        -   Taggart, R. (2022). Evaluation of point forecasts for extreme events using 
            consistent scoring functions. Quarterly Journal of the Royal Meteorological 
            Society, 148(742), 306–320. https://doi.org/10.1002/qj.4206

    """
    check_alpha(alpha)

    reduce_dims = gather_dimensions(fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims)

    score_overfcst = (1 - alpha) * (phi(obs) - phi(fcst) - phi_prime(fcst) * (obs - fcst))
    score_underfcst = alpha * (phi(obs) - phi(fcst) - phi_prime(fcst) * (obs - fcst))
    result = score_overfcst.where(obs < fcst, score_underfcst)
    result = apply_weights(result, weights=weights)
    result = result.mean(dim=reduce_dims)

    return result


def consistent_huber_score(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    huber_param: float,
    phi: Callable[[xr.DataArray], xr.DataArray],
    phi_prime: Callable[[xr.DataArray], xr.DataArray],
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[xr.DataArray] = None,
) -> xr.DataArray:
    """
    Calculates the score that is consistent for the Huber mean functional with tuning
    parameter `tuning_param`, based on convex function phi. See Taggart (2022a), or
    Equation (11) from Taggart (2022b). See Taggart (2022b), end of Section 3.4, for the
    standard formula.

    .. math::
        S(x,y) = \\frac{1}{2}(\\phi(y)-\\phi(\\kappa_v (x-y) + y) + \\kappa_v (x-y)\\phi'(x))

    where
        - :math:`\\kappa_v(x) = \\mathrm{max}(-v, \\mathrm{min}(x, v))` is the "capping function"
        - :math:`v` is the Huber parameter
        - :math:`x` is the forecast
        - :math:`y` is the observation
        - :math:`\\phi` is a convex function of a single variable
        - :math:`\\phi'` is the subderivative of :math:`\\phi`
        - :math:`S(x,y)` is the score.

    The score is negatively oriented with :math:`0 \\leq S(x,y) < \\infty`.

    We strongly recommend that you work through the tutorial
    https://scores.readthedocs.io/en/stable/tutorials/Consistent_Scores.html to help you
    understand how to use the consistent scoring functions in ``scores``.

    Args:
        fcst: array of forecast values.
        obs: array of corresponding observation values.
        huber_param: Huber mean tuning parameter. This corresponds to the transition point between
            linear and quadratic loss for Huber loss. Must be positive.
        phi: a convex function on the real numbers, accepting a single array like argument.
        phi_prime: a subderivative of `phi`, accepting a single array like argument.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the consistent Huber score. All other dimensions will be preserved. As a
            special case, 'all' will allow all dimensions to be reduced. Only one
            of `reduce_dims` and `preserve_dims` can be supplied. The default behaviour
            if neither are supplied is to reduce all dims.
        preserve_dims: Optionally specify which dimensions to preserve when calculating
            the consistent Huber score. All other dimensions will be reduced. As a special case, 'all'
            will allow all dimensions to be preserved. In this case, the result will be in
            the same shape/dimensionality as the forecast, and the errors will be the consistent Huber
            score at each point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely. Only one of `reduce_dims`
            and `preserve_dims` can be supplied. The default behaviour if neither are supplied
            is to reduce all dims.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)

    Returns:
        array of (mean) scores that is consistent for Huber mean functional,
        with the dimensions specified by `dims`. If `dims` is `None`, the returned DataArray will have
        only one entry, the overall mean score.

    Raises:
       ValueError: if `huber_param <= 0`.

    References:
        -   Taggart, R. J. (2022a). Point forecasting and forecast evaluation with
            generalized Huber loss. Electronic Journal of Statistics, 16(1), 201-231.
            https://doi.org/10.1214/21-ejs1957
        -   Taggart, R. (2022b). Evaluation of point forecasts for extreme events using
            consistent scoring functions. Quarterly Journal of the Royal Meteorological
            Society, 148(742), 306–320. https://doi.org/10.1002/qj.4206
    """
    check_huber_param(huber_param)
    reduce_dims = gather_dimensions(fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims)

    kappa = (fcst - obs).clip(min=-huber_param, max=huber_param)
    result = 0.5 * (phi(obs) - phi(kappa + obs) + kappa * phi_prime(fcst))
    result = apply_weights(result, weights=weights)
    result = result.mean(dim=reduce_dims)

    return result


def consistent_quantile_score(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    alpha: float,
    g: Callable[[xr.DataArray], xr.DataArray],
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[xr.DataArray] = None,
) -> xr.DataArray:
    """
    Calculates the score that is consistent for the alpha-quantile functional, based on 
    nondecreasing function g. See Gneiting (2011), or Equation (8) from Taggart (2022).

    .. math::

        S(x, y) =
        \\begin{cases}
        (1 - \\alpha)(g(x) - g(y)), & y < x \\\\
        \\alpha(g(y) - g(x)), & x \\leq y
        \\end{cases}

    where
        - :math:`x` is the forecast
        - :math:`y` is the observation
        - :math:`\\alpha` is the quantile level
        - :math:`g` is a nondecreasing function of a single variable
        - :math:`S(x,y)` is the score.

    The score is negatively oriented with :math:`0 \\leq S(x,y) < \\infty`.

    We strongly recommend that you work through the tutorial 
    https://scores.readthedocs.io/en/stable/tutorials/Consistent_Scores.html to help you 
    understand how to use the consistent scoring functions in ``scores``.

    Args:
        fcst: array of forecast values.
        obs: array of corresponding observation values.
        alpha: quantile level. Must be strictly between 0 and 1.
        g: nondecreasing function on the real numbers, accepting a single array like argument.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the consistent quantile score. All other dimensions will be preserved. As a
            special case, 'all' will allow all dimensions to be reduced. Only one
            of `reduce_dims` and `preserve_dims` can be supplied. The default behaviour
            if neither are supplied is to reduce all dims.
        preserve_dims: Optionally specify which dimensions to preserve when calculating
            the consistent quantile score. All other dimensions will be reduced. As a special case, 'all'
            will allow all dimensions to be preserved. In this case, the result will be in
            the same shape/dimensionality as the forecast, and the errors will be the consistent quantile
            score at each point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely. Only one of `reduce_dims`
            and `preserve_dims` can be supplied. The default behaviour if neither are supplied
            is to reduce all dims.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)

    Returns:
        array of (mean) scores that are consistent for alpha-quantile functional,
        with the dimensions specified by `dims`. If `dims` is `None`, the returned DataArray will have
        only one entry, the overall mean score.

    Raises:
        ValueError: if `alpha` is not strictly between 0 and 1.

    References:
        -   Gneiting, T. (2011). Making and Evaluating Point Forecasts. Journal of the 
            American Statistical Association, 106(494), 746–762. 
            https://doi.org/10.1198/jasa.2011.r10138
        -   Taggart, R. (2022). Evaluation of point forecasts for extreme events using 
            consistent scoring functions. Quarterly Journal of the Royal Meteorological 
            Society, 148(742), 306–320. https://doi.org/10.1002/qj.4206
    """
    check_alpha(alpha)
    reduce_dims = gather_dimensions(fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims)

    score_overfcst = (1 - alpha) * (g(fcst) - g(obs))
    score_underfcst = -alpha * (g(fcst) - g(obs))
    result = score_overfcst.where(obs < fcst, score_underfcst)
    result = apply_weights(result, weights=weights)
    result = result.mean(dim=reduce_dims)

    return result


def check_alpha(alpha: float) -> None:
    """Raises if quantile or expectile level `alpha` not in the open interval (0,1)."""
    if alpha <= 0 or alpha >= 1:
        raise ValueError("`alpha` must be strictly between 0 and 1")


def check_huber_param(huber_param: float) -> None:
    """Raises if `huber_param` is not positive."""
    if huber_param <= 0:
        raise ValueError("`huber_param` must be positive")
